from factory import DjangoModelFactory as FactoryDjangoModelFactory, Faker, SubFactory

from ..caluma_core.factories import DjangoModelFactory
from ..caluma_form.factories import DocumentFactory, FormFactory
from . import models


class TaskFactory(DjangoModelFactory):
    slug = Faker("slug")
    name = Faker("multilang", faker_provider="name")
    type = Faker("word", ext_word_list=models.Task.TYPE_CHOICES)
    address_groups = None
    control_groups = None
    description = Faker("multilang", faker_provider="text")
    meta = {}
    is_archived = False
    form = SubFactory(FormFactory)
    lead_time = None
    is_multiple_instance = False

    class Meta:
        model = models.Task


class WorkflowFactory(DjangoModelFactory):
    slug = Faker("slug")
    name = Faker("multilang", faker_provider="name")
    description = Faker("multilang", faker_provider="text")
    meta = {}
    is_published = False
    is_archived = False
    allow_all_forms = False

    class Meta:
        model = models.Workflow


class WorkflowStartTasksFactory(FactoryDjangoModelFactory):
    workflow = SubFactory(WorkflowFactory)
    task = SubFactory(TaskFactory)

    class Meta:
        model = models.Workflow.start_tasks.through


class WorkflowAllowFormsFactory(FactoryDjangoModelFactory):
    form = SubFactory(FormFactory)
    workflow = SubFactory(WorkflowFactory)

    class Meta:
        model = models.Workflow.allow_forms.through


class FlowFactory(DjangoModelFactory):
    next = Faker("slug")

    class Meta:
        model = models.Flow


class TaskFlowFactory(DjangoModelFactory):
    workflow = SubFactory(WorkflowFactory)
    task = SubFactory(TaskFactory)
    flow = SubFactory(FlowFactory)

    class Meta:
        model = models.TaskFlow


class CaseFactory(DjangoModelFactory):
    workflow = SubFactory(WorkflowFactory)
    status = models.Case.STATUS_RUNNING
    meta = {}
    document = SubFactory(DocumentFactory)

    class Meta:
        model = models.Case


class WorkItemFactory(DjangoModelFactory):
    case = SubFactory(CaseFactory)
    child_case = SubFactory(CaseFactory)
    task = SubFactory(TaskFactory)
    status = models.WorkItem.STATUS_READY
    document = SubFactory(DocumentFactory)
    addressed_groups = []
    assigned_users = []
    controlling_groups = []
    meta = {}
    deadline = None

    class Meta:
        model = models.WorkItem



from collections import namedtuple

from django.conf import settings
from django.utils.module_loading import import_string
from localized_fields.value import LocalizedValue

from caluma.utils import is_iterable_and_no_string

DataSource = namedtuple("DataSource", ["name", "info"])


class DataSourceException(Exception):
    pass


class Data:
    def __init__(self, data):
        self.data = data
        self.slug, self.label = self.load()

    def load(self):
        if is_iterable_and_no_string(self.data):
            if not 0 < len(self.data) < 3:
                raise DataSourceException(f"Failed to parse data:\n{self.data}")
            elif len(self.data) == 1:
                return str(self.data[0]), str(self.data[0])
            elif isinstance(self.data[1], dict):
                label = LocalizedValue(self.data[1])
                return (str(self.data[0]), str(label.translate()))
            return str(self.data[0]), str(self.data[1])

        return str(self.data), str(self.data)


def get_data_sources(dic=False):
    """Get all configured DataSources.

    :param dic: Should return a dict
    :return: List of DataSource-objects if dic False otherwise dict
    """

    data_source_classes = [import_string(cls) for cls in settings.DATA_SOURCE_CLASSES]
    if dic:
        return {ds.__name__: ds for ds in data_source_classes}
    return [
        DataSource(
            name=ds.__name__,
            info=LocalizedValue(ds.info).translate()
            if isinstance(ds.info, dict)
            else ds.info,
        )
        for ds in data_source_classes
    ]


def get_data_source_data(info, name):
    data_sources = get_data_sources(dic=True)
    if name not in data_sources:
        raise DataSourceException(f"No data_source found for name: {name}")

    raw_data = data_sources[name]().try_get_data_with_fallback(info)
    if not is_iterable_and_no_string(raw_data):
        raise DataSourceException(f"Failed to parse data from source: {name}")

    return [Data(d) for d in raw_data]



from graphql_relay.connection.arrayconnection import (
    get_offset_with_default,
    offset_to_cursor,
)
from graphql_relay.connection.connectiontypes import Connection, Edge, PageInfo


def connection_from_list(data, args=None, **kwargs):
    """
    Replace graphql_relay.connection.arrayconnection.connection_from_list.

    This can be removed, when (or better if)
    https://github.com/graphql-python/graphql-relay-py/issues/12
    is resolved.

    A simple function that accepts an array and connection arguments, and returns
    a connection object for use in GraphQL. It uses array offsets as pagination,
    so pagination will only work if the array is static.
    """
    _len = len(data)
    return connection_from_list_slice(
        data, args, slice_start=0, list_length=_len, list_slice_length=_len, **kwargs
    )


def connection_from_list_slice(
    list_slice,
    args=None,
    connection_type=None,
    edge_type=None,
    pageinfo_type=None,
    slice_start=0,
    list_length=0,
    list_slice_length=None,
):
    """
    Replace graphql_relay.connection.arrayconnection.connection_from_list_slice.

    This can be removed, when (or better if)
    https://github.com/graphql-python/graphql-relay-py/issues/12
    is resolved.

    Given a slice (subset) of an array, returns a connection object for use in
    GraphQL.
    This function is similar to `connectionFromArray`, but is intended for use
    cases where you know the cardinality of the connection, consider it too large
    to materialize the entire array, and instead wish pass in a slice of the
    total result large enough to cover the range specified in `args`.
    """
    connection_type = connection_type or Connection
    edge_type = edge_type or Edge
    pageinfo_type = pageinfo_type or PageInfo

    args = args or {}

    before = args.get("before")
    after = args.get("after")
    first = args.get("first")
    last = args.get("last")
    if list_slice_length is None:  # pragma: no cover
        list_slice_length = len(list_slice)
    slice_end = slice_start + list_slice_length
    before_offset = get_offset_with_default(before, list_length)
    after_offset = get_offset_with_default(after, -1)

    start_offset = max(slice_start - 1, after_offset, -1) + 1
    end_offset = min(slice_end, before_offset, list_length)
    if isinstance(first, int):
        end_offset = min(end_offset, start_offset + first)
    if isinstance(last, int):
        start_offset = max(start_offset, end_offset - last)

    # If supplied slice is too large, trim it down before mapping over it.
    _slice = list_slice[
        max(start_offset - slice_start, 0) : list_slice_length
        - (slice_end - end_offset)
    ]
    edges = [
        edge_type(node=node, cursor=offset_to_cursor(start_offset + i))
        for i, node in enumerate(_slice)
    ]

    first_edge_cursor = edges[0].cursor if edges else None
    last_edge_cursor = edges[-1].cursor if edges else None

    return connection_type(
        edges=edges,
        page_info=pageinfo_type(
            start_cursor=first_edge_cursor,
            end_cursor=last_edge_cursor,
            has_previous_page=start_offset > 0,
            has_next_page=end_offset < list_length,
        ),
    )



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots[
    "test_schema_introspect 1"
] = """schema {
  query: Query
  mutation: Mutation
}

input AddFormQuestionInput {
  form: ID!
  question: ID!
  clientMutationId: String
}

type AddFormQuestionPayload {
  form: Form
  clientMutationId: String
}

input AddWorkflowFlowInput {
  workflow: ID!
  tasks: [ID]!
  next: FlowJexl!
  clientMutationId: String
}

type AddWorkflowFlowPayload {
  workflow: Workflow
  clientMutationId: String
}

interface Answer {
  id: ID
  createdAt: DateTime!
  createdByUser: String
  createdByGroup: String
  modifiedAt: DateTime!
  question: Question!
  meta: GenericScalar!
}

type AnswerConnection {
  pageInfo: PageInfo!
  edges: [AnswerEdge]!
  totalCount: Int
}

type AnswerEdge {
  node: Answer
  cursor: String!
}

input AnswerFilterSetType {
  question: ID
  search: String
  createdByUser: String
  createdByGroup: String
  createdBefore: DateTime
  createdAfter: DateTime
  metaHasKey: String
  metaValue: [JSONValueFilterType]
  orderBy: [AnswerOrdering]
  questions: [ID]
  visibleInContext: Boolean
  invert: Boolean
}

enum AnswerHierarchyMode {
  DIRECT
  FAMILY
}

enum AnswerLookupMode {
  EXACT
  STARTSWITH
  CONTAINS
  ICONTAINS
  INTERSECTS
  ISNULL
  GTE
  GT
  LTE
  LT
}

input AnswerOrderSetType {
  meta: String
  attribute: SortableAnswerAttributes
  direction: AscDesc
}

enum AnswerOrdering {
  CREATED_AT_ASC
  CREATED_AT_DESC
  MODIFIED_AT_ASC
  MODIFIED_AT_DESC
  CREATED_BY_USER_ASC
  CREATED_BY_USER_DESC
  CREATED_BY_GROUP_ASC
  CREATED_BY_GROUP_DESC
  META_TEST_KEY_ASC
  META_TEST_KEY_DESC
  META_FOOBAR_ASC
  META_FOOBAR_DESC
}

enum AscDesc {
  ASC
  DESC
}

input CancelCaseInput {
  id: ID!
  clientMutationId: String
}

type CancelCasePayload {
  case: Case
  clientMutationId: String
}

type Case implements Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  id: ID!
  closedAt: DateTime
  closedByUser: String
  closedByGroup: String
  workflow: Workflow!
  status: CaseStatus!
  meta: GenericScalar
  document: Document
  workItems(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], status: WorkItemStatusArgument, name: String, task: ID, case: ID, createdAt: DateTime, closedAt: DateTime, modifiedAt: DateTime, orderBy: [WorkItemOrdering], filter: [WorkItemFilterSetType], order: [WorkItemOrderSetType], createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, addressedGroups: [String], controllingGroups: [String], assignedUsers: [String], documentHasAnswer: [HasAnswerFilterType], caseDocumentHasAnswer: [HasAnswerFilterType], caseMetaValue: [JSONValueFilterType]): WorkItemConnection
  parentWorkItem: WorkItem
  familyWorkItems(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], status: WorkItemStatusArgument, orderBy: [WorkItemOrdering], filter: [WorkItemFilterSetType], order: [WorkItemOrderSetType], documentHasAnswer: [HasAnswerFilterType], caseDocumentHasAnswer: [HasAnswerFilterType], caseMetaValue: [JSONValueFilterType], name: String, task: ID, case: ID, createdAt: DateTime, closedAt: DateTime, modifiedAt: DateTime, createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, addressedGroups: [String], controllingGroups: [String], assignedUsers: [String]): WorkItemConnection
}

type CaseConnection {
  pageInfo: PageInfo!
  edges: [CaseEdge]!
  totalCount: Int
}

type CaseEdge {
  node: Case
  cursor: String!
}

input CaseFilterSetType {
  workflow: ID
  createdByUser: String
  createdByGroup: String
  createdBefore: DateTime
  createdAfter: DateTime
  metaHasKey: String
  metaValue: [JSONValueFilterType]
  orderBy: [CaseOrdering]
  documentForm: String
  documentForms: [String]
  hasAnswer: [HasAnswerFilterType]
  workItemDocumentHasAnswer: [HasAnswerFilterType]
  rootCase: ID
  searchAnswers: [SearchAnswersFilterType]
  status: [CaseStatusArgument]
  orderByQuestionAnswerValue: String
  invert: Boolean
}

input CaseOrderSetType {
  meta: String
  attribute: SortableCaseAttributes
  documentAnswer: String
  direction: AscDesc
}

enum CaseOrdering {
  STATUS_ASC
  STATUS_DESC
  CREATED_AT_ASC
  CREATED_AT_DESC
  MODIFIED_AT_ASC
  MODIFIED_AT_DESC
  CREATED_BY_USER_ASC
  CREATED_BY_USER_DESC
  CREATED_BY_GROUP_ASC
  CREATED_BY_GROUP_DESC
  META_TEST_KEY_ASC
  META_TEST_KEY_DESC
  META_FOOBAR_ASC
  META_FOOBAR_DESC
}

enum CaseStatus {
  RUNNING
  COMPLETED
  CANCELED
}

enum CaseStatusArgument {
  RUNNING
  COMPLETED
  CANCELED
}

type ChoiceQuestion implements Question, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  slug: String!
  label: String!
  isRequired: QuestionJexl!
  isHidden: QuestionJexl!
  isArchived: Boolean!
  infoText: String
  meta: GenericScalar!
  source: Question
  forms(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], orderBy: [FormOrdering], slug: String, name: String, description: String, isPublished: Boolean, isArchived: Boolean, createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, search: String, slugs: [String]): FormConnection
  options(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], slug: String, label: String, isArchived: Boolean, createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, search: String, orderBy: [OptionOrdering]): OptionConnection
  id: ID!
}

type CompleteTaskFormTask implements Task, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  slug: String!
  name: String!
  description: String
  type: TaskType!
  meta: GenericScalar!
  addressGroups: GroupJexl
  controlGroups: GroupJexl
  isArchived: Boolean!
  leadTime: Int
  isMultipleInstance: Boolean!
  form: Form!
  id: ID!
}

input CompleteWorkItemInput {
  id: ID!
  clientMutationId: String
}

type CompleteWorkItemPayload {
  workItem: WorkItem
  clientMutationId: String
}

type CompleteWorkflowFormTask implements Task, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  slug: String!
  name: String!
  description: String
  type: TaskType!
  meta: GenericScalar!
  addressGroups: GroupJexl
  controlGroups: GroupJexl
  isArchived: Boolean!
  leadTime: Int
  isMultipleInstance: Boolean!
  id: ID!
}

input CopyDocumentInput {
  source: ID!
  clientMutationId: String
}

type CopyDocumentPayload {
  document: Document
  clientMutationId: String
}

input CopyFormInput {
  slug: String!
  name: String!
  description: String
  source: ID!
  isPublished: Boolean
  clientMutationId: String
}

type CopyFormPayload {
  form: Form
  clientMutationId: String
}

input CopyOptionInput {
  slug: String!
  label: String!
  source: ID!
  clientMutationId: String
}

type CopyOptionPayload {
  option: Option
  clientMutationId: String
}

input CopyQuestionInput {
  slug: String!
  label: String!
  source: ID!
  clientMutationId: String
}

type CopyQuestionPayload {
  question: Question
  clientMutationId: String
}

input CreateWorkItemInput {
  case: ID!
  multipleInstanceTask: ID!
  name: String
  description: String
  assignedUsers: [String]
  addressedGroups: [String]
  controllingGroups: [String]
  deadline: DateTime
  meta: JSONString
  clientMutationId: String
}

type CreateWorkItemPayload {
  workItem: WorkItem
  clientMutationId: String
}

type DataSource {
  info: String
  name: String!
}

type DataSourceConnection {
  pageInfo: PageInfo!
  edges: [DataSourceEdge]!
  totalCount: Int
}

type DataSourceData {
  label: String!
  slug: String!
}

type DataSourceDataConnection {
  pageInfo: PageInfo!
  edges: [DataSourceDataEdge]!
  totalCount: Int
}

type DataSourceDataEdge {
  node: DataSourceData
  cursor: String!
}

type DataSourceEdge {
  node: DataSource
  cursor: String!
}

scalar Date

type DateAnswer implements Answer, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  id: ID!
  question: Question!
  value: Date
  meta: GenericScalar!
  date: Date
}

type DateQuestion implements Question, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  slug: String!
  label: String!
  isRequired: QuestionJexl!
  isHidden: QuestionJexl!
  isArchived: Boolean!
  infoText: String
  meta: GenericScalar!
  source: Question
  forms(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], orderBy: [FormOrdering], slug: String, name: String, description: String, isPublished: Boolean, isArchived: Boolean, createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, search: String, slugs: [String]): FormConnection
  id: ID!
}

scalar DateTime

type DjangoDebug {
  sql: [DjangoDebugSQL]
}

type DjangoDebugSQL {
  vendor: String!
  alias: String!
  sql: String
  duration: Float!
  rawSql: String!
  params: String!
  startTime: Float!
  stopTime: Float!
  isSlow: Boolean!
  isSelect: Boolean!
  transId: String
  transStatus: String
  isoLevel: String
  encoding: String
}

type Document implements Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  id: ID!
  form: Form!
  source: Document
  meta: GenericScalar
  copies(before: String, after: String, first: Int, last: Int): DocumentConnection!
  answers(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], question: ID, search: String, orderBy: [AnswerOrdering], filter: [AnswerFilterSetType], order: [AnswerOrderSetType], createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, questions: [ID], visibleInContext: Boolean): AnswerConnection
  case: Case
  workItem: WorkItem
}

type DocumentConnection {
  pageInfo: PageInfo!
  edges: [DocumentEdge]!
  totalCount: Int
}

type DocumentEdge {
  node: Document
  cursor: String!
}

input DocumentFilterSetType {
  form: ID
  forms: [ID]
  search: String
  id: ID
  createdByUser: String
  createdByGroup: String
  createdBefore: DateTime
  createdAfter: DateTime
  metaHasKey: String
  metaValue: [JSONValueFilterType]
  orderBy: [DocumentOrdering]
  rootDocument: ID
  hasAnswer: [HasAnswerFilterType]
  searchAnswers: [SearchAnswersFilterType]
  invert: Boolean
}

input DocumentOrderSetType {
  meta: String
  answerValue: String
  attribute: SortableDocumentAttributes
  direction: AscDesc
}

enum DocumentOrdering {
  CREATED_AT_ASC
  CREATED_AT_DESC
  MODIFIED_AT_ASC
  MODIFIED_AT_DESC
  CREATED_BY_USER_ASC
  CREATED_BY_USER_DESC
  CREATED_BY_GROUP_ASC
  CREATED_BY_GROUP_DESC
  META_TEST_KEY_ASC
  META_TEST_KEY_DESC
  META_FOOBAR_ASC
  META_FOOBAR_DESC
}

type DocumentValidityConnection {
  pageInfo: PageInfo!
  edges: [DocumentValidityEdge]!
  totalCount: Int
}

type DocumentValidityEdge {
  node: ValidationResult
  cursor: String!
}

type DynamicChoiceQuestion implements Question, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  slug: String!
  label: String!
  isRequired: QuestionJexl!
  isHidden: QuestionJexl!
  isArchived: Boolean!
  infoText: String
  meta: GenericScalar!
  source: Question
  forms(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], orderBy: [FormOrdering], slug: String, name: String, description: String, isPublished: Boolean, isArchived: Boolean, createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, search: String, slugs: [String]): FormConnection
  options(before: String, after: String, first: Int, last: Int): DataSourceDataConnection
  dataSource: String!
  id: ID!
}

type DynamicMultipleChoiceQuestion implements Question, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  slug: String!
  label: String!
  isRequired: QuestionJexl!
  isHidden: QuestionJexl!
  isArchived: Boolean!
  infoText: String
  meta: GenericScalar!
  source: Question
  forms(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], orderBy: [FormOrdering], slug: String, name: String, description: String, isPublished: Boolean, isArchived: Boolean, createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, search: String, slugs: [String]): FormConnection
  options(before: String, after: String, first: Int, last: Int): DataSourceDataConnection
  dataSource: String!
  id: ID!
}

type DynamicOption implements Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  id: ID!
  slug: String!
  label: String!
  document: Document!
  question: StaticQuestion!
}

type DynamicOptionConnection {
  pageInfo: PageInfo!
  edges: [DynamicOptionEdge]!
  totalCount: Int
}

type DynamicOptionEdge {
  node: DynamicOption
  cursor: String!
}

input DynamicOptionFilterSetType {
  question: ID
  document: ID
  invert: Boolean
}

type File implements Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  id: ID!
  name: String!
  answer: FileAnswer
  uploadUrl: String
  downloadUrl: String
  metadata: GenericScalar
}

type FileAnswer implements Answer, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  id: ID!
  question: Question!
  value: File!
  meta: GenericScalar!
  file: File
}

type FileQuestion implements Question, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  slug: String!
  label: String!
  isRequired: QuestionJexl!
  isHidden: QuestionJexl!
  isArchived: Boolean!
  infoText: String
  meta: GenericScalar!
  source: Question
  forms(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], orderBy: [FormOrdering], slug: String, name: String, description: String, isPublished: Boolean, isArchived: Boolean, createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, search: String, slugs: [String]): FormConnection
  id: ID!
}

type FloatAnswer implements Answer, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  id: ID!
  question: Question!
  value: Float
  meta: GenericScalar!
}

type FloatQuestion implements Question, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  slug: String!
  label: String!
  isRequired: QuestionJexl!
  isHidden: QuestionJexl!
  isArchived: Boolean!
  placeholder: String
  infoText: String
  meta: GenericScalar!
  source: Question
  forms(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], orderBy: [FormOrdering], slug: String, name: String, description: String, isPublished: Boolean, isArchived: Boolean, createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, search: String, slugs: [String]): FormConnection
  id: ID!
  minValue: Float
  maxValue: Float
}

type Flow implements Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  id: ID!
  next: FlowJexl!
  tasks: [Task]!
}

type FlowConnection {
  pageInfo: PageInfo!
  edges: [FlowEdge]!
  totalCount: Int
}

type FlowEdge {
  node: Flow
  cursor: String!
}

input FlowFilterSetType {
  task: ID
  createdByUser: String
  createdByGroup: String
  createdBefore: DateTime
  createdAfter: DateTime
  invert: Boolean
}

scalar FlowJexl

type Form implements Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  slug: String!
  name: String!
  description: String
  meta: GenericScalar
  isPublished: Boolean!
  isArchived: Boolean!
  questions(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], slug: String, label: String, isRequired: String, isHidden: String, isArchived: Boolean, createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, excludeForms: [ID], search: String, orderBy: [QuestionOrdering], slugs: [String]): QuestionConnection
  source: Form
  documents(before: String, after: String, first: Int, last: Int): DocumentConnection!
  id: ID!
}

type FormConnection {
  pageInfo: PageInfo!
  edges: [FormEdge]!
  totalCount: Int
}

type FormEdge {
  node: Form
  cursor: String!
}

input FormFilterSetType {
  orderBy: [FormOrdering]
  slug: String
  name: String
  description: String
  isPublished: Boolean
  isArchived: Boolean
  createdByUser: String
  createdByGroup: String
  createdBefore: DateTime
  createdAfter: DateTime
  metaHasKey: String
  metaValue: [JSONValueFilterType]
  search: String
  slugs: [String]
  invert: Boolean
}

input FormOrderSetType {
  meta: String
  attribute: SortableFormAttributes
  direction: AscDesc
}

enum FormOrdering {
  NAME_ASC
  NAME_DESC
  CREATED_AT_ASC
  CREATED_AT_DESC
  MODIFIED_AT_ASC
  MODIFIED_AT_DESC
  CREATED_BY_USER_ASC
  CREATED_BY_USER_DESC
  CREATED_BY_GROUP_ASC
  CREATED_BY_GROUP_DESC
  META_TEST_KEY_ASC
  META_TEST_KEY_DESC
  META_FOOBAR_ASC
  META_FOOBAR_DESC
}

type FormQuestion implements Question, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  slug: String!
  label: String!
  isRequired: QuestionJexl!
  isHidden: QuestionJexl!
  isArchived: Boolean!
  infoText: String
  meta: GenericScalar!
  source: Question
  forms(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], orderBy: [FormOrdering], slug: String, name: String, description: String, isPublished: Boolean, isArchived: Boolean, createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, search: String, slugs: [String]): FormConnection
  subForm: Form
  id: ID!
}

type FormatValidator {
  slug: String!
  name: String!
  regex: String!
  errorMsg: String!
}

type FormatValidatorConnection {
  pageInfo: PageInfo!
  edges: [FormatValidatorEdge]!
  totalCount: Int
}

type FormatValidatorEdge {
  node: FormatValidator
  cursor: String!
}

scalar GenericScalar

scalar GroupJexl

input HasAnswerFilterType {
  question: ID!
  value: GenericScalar
  lookup: AnswerLookupMode
  hierarchy: AnswerHierarchyMode
}

interface HistoricalAnswer {
  id: ID
  createdAt: DateTime!
  createdByUser: String
  createdByGroup: String
  modifiedAt: DateTime!
  question: Question!
  meta: GenericScalar!
  historyDate: DateTime!
  historyUserId: String
  historyType: String
}

type HistoricalAnswerConnection {
  pageInfo: PageInfo!
  edges: [HistoricalAnswerEdge]!
  totalCount: Int
}

type HistoricalAnswerEdge {
  node: HistoricalAnswer
  cursor: String!
}

type HistoricalDateAnswer implements HistoricalAnswer, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  id: ID!
  value: Date
  meta: GenericScalar!
  date: Date
  historyQuestionType: String!
  historyUserId: String
  question: Question!
  historyId: UUID!
  historyDate: DateTime!
  historyChangeReason: String
  historyType: String
}

type HistoricalDocument implements Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  id: ID!
  meta: GenericScalar
  historyUserId: String
  form: Form
  source: Document
  historyDate: DateTime!
  historyType: String
  historicalAnswers(asOf: DateTime!, before: String, after: String, first: Int, last: Int): HistoricalAnswerConnection
  documentId: UUID
}

type HistoricalFile implements Node {
  id: ID!
  name: String!
  downloadUrl: String
  metadata: GenericScalar
  historicalAnswer: HistoricalFileAnswer
  historyDate: DateTime!
  historyUserId: String
  historyType: String
}

type HistoricalFileAnswer implements HistoricalAnswer, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  id: ID!
  value(asOf: DateTime!): HistoricalFile
  meta: GenericScalar!
  historyQuestionType: String!
  historyUserId: String
  question: Question!
  historyId: UUID!
  historyDate: DateTime!
  historyChangeReason: String
  historyType: String
  file: File
}

type HistoricalFloatAnswer implements HistoricalAnswer, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  id: ID!
  value: Float
  meta: GenericScalar!
  historyQuestionType: String!
  historyUserId: String
  question: Question!
  historyId: UUID!
  historyDate: DateTime!
  historyChangeReason: String
  historyType: String
}

type HistoricalIntegerAnswer implements HistoricalAnswer, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  id: ID!
  value: Int
  meta: GenericScalar!
  historyQuestionType: String!
  historyUserId: String
  question: Question!
  historyId: UUID!
  historyDate: DateTime!
  historyChangeReason: String
  historyType: String
}

type HistoricalListAnswer implements HistoricalAnswer, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  id: ID!
  value: [String]
  meta: GenericScalar!
  historyQuestionType: String!
  historyUserId: String
  question: Question!
  historyId: UUID!
  historyDate: DateTime!
  historyChangeReason: String
  historyType: String
}

type HistoricalStringAnswer implements HistoricalAnswer, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  id: ID!
  value: String
  meta: GenericScalar!
  historyQuestionType: String!
  historyUserId: String
  question: Question!
  historyId: UUID!
  historyDate: DateTime!
  historyChangeReason: String
  historyType: String
}

type HistoricalTableAnswer implements HistoricalAnswer, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  id: ID!
  value(asOf: DateTime!): [HistoricalDocument]
  meta: GenericScalar!
  historyQuestionType: String!
  historyUserId: String
  question: Question!
  historyId: UUID!
  historyDate: DateTime!
  historyChangeReason: String
  historyType: String
  document: Document
}

type IntegerAnswer implements Answer, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  id: ID!
  question: Question!
  value: Int
  meta: GenericScalar!
}

type IntegerQuestion implements Question, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  slug: String!
  label: String!
  isRequired: QuestionJexl!
  isHidden: QuestionJexl!
  isArchived: Boolean!
  placeholder: String
  infoText: String
  meta: GenericScalar!
  source: Question
  forms(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], orderBy: [FormOrdering], slug: String, name: String, description: String, isPublished: Boolean, isArchived: Boolean, createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, search: String, slugs: [String]): FormConnection
  id: ID!
  maxValue: Int
  minValue: Int
}

enum JSONLookupMode {
  EXACT
  STARTSWITH
  CONTAINS
  ICONTAINS
  GTE
  GT
  LTE
  LT
}

scalar JSONString

input JSONValueFilterType {
  key: String!
  value: GenericScalar!
  lookup: JSONLookupMode
}

type ListAnswer implements Answer, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  id: ID!
  question: Question!
  value: [String]
  meta: GenericScalar!
}

type MultipleChoiceQuestion implements Question, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  slug: String!
  label: String!
  isRequired: QuestionJexl!
  isHidden: QuestionJexl!
  isArchived: Boolean!
  infoText: String
  meta: GenericScalar!
  source: Question
  forms(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], orderBy: [FormOrdering], slug: String, name: String, description: String, isPublished: Boolean, isArchived: Boolean, createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, search: String, slugs: [String]): FormConnection
  options(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], orderBy: [OptionOrdering], slug: String, label: String, isArchived: Boolean, createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, search: String): OptionConnection
  staticContent: String
  id: ID!
}

type Mutation {
  saveWorkflow(input: SaveWorkflowInput!): SaveWorkflowPayload
  addWorkflowFlow(input: AddWorkflowFlowInput!): AddWorkflowFlowPayload
  removeFlow(input: RemoveFlowInput!): RemoveFlowPayload
  saveSimpleTask(input: SaveSimpleTaskInput!): SaveSimpleTaskPayload
  saveCompleteWorkflowFormTask(input: SaveCompleteWorkflowFormTaskInput!): SaveCompleteWorkflowFormTaskPayload
  saveCompleteTaskFormTask(input: SaveCompleteTaskFormTaskInput!): SaveCompleteTaskFormTaskPayload
  startCase(input: StartCaseInput!): StartCasePayload @deprecated(reason: "Use SaveCase mutation instead")
  saveCase(input: SaveCaseInput!): SaveCasePayload
  cancelCase(input: CancelCaseInput!): CancelCasePayload
  completeWorkItem(input: CompleteWorkItemInput!): CompleteWorkItemPayload
  skipWorkItem(input: SkipWorkItemInput!): SkipWorkItemPayload
  saveWorkItem(input: SaveWorkItemInput!): SaveWorkItemPayload
  createWorkItem(input: CreateWorkItemInput!): CreateWorkItemPayload
  saveForm(input: SaveFormInput!): SaveFormPayload
  copyForm(input: CopyFormInput!): CopyFormPayload
  addFormQuestion(input: AddFormQuestionInput!): AddFormQuestionPayload
  removeFormQuestion(input: RemoveFormQuestionInput!): RemoveFormQuestionPayload
  reorderFormQuestions(input: ReorderFormQuestionsInput!): ReorderFormQuestionsPayload
  saveOption(input: SaveOptionInput!): SaveOptionPayload
  copyOption(input: CopyOptionInput!): CopyOptionPayload
  copyQuestion(input: CopyQuestionInput!): CopyQuestionPayload
  saveTextQuestion(input: SaveTextQuestionInput!): SaveTextQuestionPayload
  saveTextareaQuestion(input: SaveTextareaQuestionInput!): SaveTextareaQuestionPayload
  saveDateQuestion(input: SaveDateQuestionInput!): SaveDateQuestionPayload
  saveChoiceQuestion(input: SaveChoiceQuestionInput!): SaveChoiceQuestionPayload
  saveMultipleChoiceQuestion(input: SaveMultipleChoiceQuestionInput!): SaveMultipleChoiceQuestionPayload
  saveDynamicChoiceQuestion(input: SaveDynamicChoiceQuestionInput!): SaveDynamicChoiceQuestionPayload
  saveDynamicMultipleChoiceQuestion(input: SaveDynamicMultipleChoiceQuestionInput!): SaveDynamicMultipleChoiceQuestionPayload
  saveFloatQuestion(input: SaveFloatQuestionInput!): SaveFloatQuestionPayload
  saveIntegerQuestion(input: SaveIntegerQuestionInput!): SaveIntegerQuestionPayload
  saveTableQuestion(input: SaveTableQuestionInput!): SaveTableQuestionPayload
  saveFormQuestion(input: SaveFormQuestionInput!): SaveFormQuestionPayload
  saveFileQuestion(input: SaveFileQuestionInput!): SaveFileQuestionPayload
  saveStaticQuestion(input: SaveStaticQuestionInput!): SaveStaticQuestionPayload
  copyDocument(input: CopyDocumentInput!): CopyDocumentPayload
  saveDocument(input: SaveDocumentInput!): SaveDocumentPayload
  saveDocumentStringAnswer(input: SaveDocumentStringAnswerInput!): SaveDocumentStringAnswerPayload
  saveDocumentIntegerAnswer(input: SaveDocumentIntegerAnswerInput!): SaveDocumentIntegerAnswerPayload
  saveDocumentFloatAnswer(input: SaveDocumentFloatAnswerInput!): SaveDocumentFloatAnswerPayload
  saveDocumentDateAnswer(input: SaveDocumentDateAnswerInput!): SaveDocumentDateAnswerPayload
  saveDocumentListAnswer(input: SaveDocumentListAnswerInput!): SaveDocumentListAnswerPayload
  saveDocumentTableAnswer(input: SaveDocumentTableAnswerInput!): SaveDocumentTableAnswerPayload
  saveDocumentFileAnswer(input: SaveDocumentFileAnswerInput!): SaveDocumentFileAnswerPayload
  removeAnswer(input: RemoveAnswerInput!): RemoveAnswerPayload
  removeDocument(input: RemoveDocumentInput!): RemoveDocumentPayload
}

interface Node {
  id: ID!
}

type Option implements Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  slug: String!
  label: String!
  isArchived: Boolean!
  meta: GenericScalar
  source: Option
  id: ID!
}

type OptionConnection {
  pageInfo: PageInfo!
  edges: [OptionEdge]!
  totalCount: Int
}

type OptionEdge {
  node: Option
  cursor: String!
}

enum OptionOrdering {
  LABEL_ASC
  LABEL_DESC
  CREATED_AT_ASC
  CREATED_AT_DESC
  MODIFIED_AT_ASC
  MODIFIED_AT_DESC
  CREATED_BY_USER_ASC
  CREATED_BY_USER_DESC
  CREATED_BY_GROUP_ASC
  CREATED_BY_GROUP_DESC
  META_TEST_KEY_ASC
  META_TEST_KEY_DESC
  META_FOOBAR_ASC
  META_FOOBAR_DESC
}

type PageInfo {
  hasNextPage: Boolean!
  hasPreviousPage: Boolean!
  startCursor: String
  endCursor: String
}

type Query {
  documentAsOf(id: ID!, asOf: DateTime!): HistoricalDocument
  allDataSources(before: String, after: String, first: Int, last: Int): DataSourceConnection
  dataSource(name: String!, before: String, after: String, first: Int, last: Int): DataSourceDataConnection
  allWorkflows(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], slug: String, name: String, description: String, isPublished: Boolean, isArchived: Boolean, orderBy: [WorkflowOrdering], filter: [WorkflowFilterSetType], order: [WorkflowOrderSetType], createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, search: String): WorkflowConnection
  allTasks(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], slug: String, name: String, description: String, type: TaskTypeArgument, isArchived: Boolean, orderBy: [TaskOrdering], filter: [TaskFilterSetType], order: [TaskOrderSetType], createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, search: String): TaskConnection
  allCases(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], workflow: ID, orderBy: [CaseOrdering], filter: [CaseFilterSetType], order: [CaseOrderSetType], createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, documentForm: String, documentForms: [String], hasAnswer: [HasAnswerFilterType], workItemDocumentHasAnswer: [HasAnswerFilterType], rootCase: ID, searchAnswers: [SearchAnswersFilterType], status: [CaseStatusArgument], orderByQuestionAnswerValue: String): CaseConnection
  allWorkItems(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], status: WorkItemStatusArgument, orderBy: [WorkItemOrdering], filter: [WorkItemFilterSetType], order: [WorkItemOrderSetType], documentHasAnswer: [HasAnswerFilterType], caseDocumentHasAnswer: [HasAnswerFilterType], caseMetaValue: [JSONValueFilterType], name: String, task: ID, case: ID, createdAt: DateTime, closedAt: DateTime, modifiedAt: DateTime, createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, addressedGroups: [String], controllingGroups: [String], assignedUsers: [String]): WorkItemConnection
  allForms(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], orderBy: [FormOrdering], slug: String, name: String, description: String, isPublished: Boolean, isArchived: Boolean, filter: [FormFilterSetType], order: [FormOrderSetType], createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, search: String, slugs: [String]): FormConnection
  allQuestions(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], orderBy: [QuestionOrdering], slug: String, label: String, isRequired: String, isHidden: String, isArchived: Boolean, filter: [QuestionFilterSetType], order: [QuestionOrderSetType], createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, excludeForms: [ID], search: String, slugs: [String]): QuestionConnection
  allDocuments(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], form: ID, forms: [ID], search: String, id: ID, orderBy: [DocumentOrdering], filter: [DocumentFilterSetType], order: [DocumentOrderSetType], createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, rootDocument: ID, hasAnswer: [HasAnswerFilterType], searchAnswers: [SearchAnswersFilterType]): DocumentConnection
  allFormatValidators(before: String, after: String, first: Int, last: Int): FormatValidatorConnection
  allUsedDynamicOptions(before: String, after: String, first: Int, last: Int, question: ID, document: ID, filter: [DynamicOptionFilterSetType], createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime): DynamicOptionConnection
  documentValidity(id: ID!, before: String, after: String, first: Int, last: Int): DocumentValidityConnection
  node(id: ID!): Node
  _debug: DjangoDebug
}

interface Question {
  id: ID!
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  slug: String!
  label: String!
  infoText: String
  isRequired: QuestionJexl!
  isHidden: QuestionJexl!
  isArchived: Boolean!
  meta: GenericScalar!
  forms(before: String, after: String, first: Int, last: Int, slug: String, name: String, description: String, isPublished: Boolean, isArchived: Boolean, createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, metaValue: [JSONValueFilterType], search: String, orderBy: [FormOrdering], slugs: [String]): FormConnection
  source: Question
}

type QuestionConnection {
  pageInfo: PageInfo!
  edges: [QuestionEdge]!
  totalCount: Int
}

type QuestionEdge {
  node: Question
  cursor: String!
}

input QuestionFilterSetType {
  orderBy: [QuestionOrdering]
  slug: String
  label: String
  isRequired: String
  isHidden: String
  isArchived: Boolean
  createdByUser: String
  createdByGroup: String
  createdBefore: DateTime
  createdAfter: DateTime
  metaHasKey: String
  metaValue: [JSONValueFilterType]
  excludeForms: [ID]
  search: String
  slugs: [String]
  invert: Boolean
}

scalar QuestionJexl

input QuestionOrderSetType {
  meta: String
  attribute: SortableQuestionAttributes
  direction: AscDesc
}

enum QuestionOrdering {
  LABEL_ASC
  LABEL_DESC
  CREATED_AT_ASC
  CREATED_AT_DESC
  MODIFIED_AT_ASC
  MODIFIED_AT_DESC
  CREATED_BY_USER_ASC
  CREATED_BY_USER_DESC
  CREATED_BY_GROUP_ASC
  CREATED_BY_GROUP_DESC
  META_TEST_KEY_ASC
  META_TEST_KEY_DESC
  META_FOOBAR_ASC
  META_FOOBAR_DESC
}

input RemoveAnswerInput {
  answer: ID!
  clientMutationId: String
}

type RemoveAnswerPayload {
  answer: Answer
  clientMutationId: String
}

input RemoveDocumentInput {
  document: ID!
  clientMutationId: String
}

type RemoveDocumentPayload {
  document: Document
  clientMutationId: String
}

input RemoveFlowInput {
  flow: ID!
  clientMutationId: String
}

type RemoveFlowPayload {
  flow: Flow
  clientMutationId: String
}

input RemoveFormQuestionInput {
  form: ID!
  question: ID!
  clientMutationId: String
}

type RemoveFormQuestionPayload {
  form: Form
  clientMutationId: String
}

input ReorderFormQuestionsInput {
  form: ID!
  questions: [ID]!
  clientMutationId: String
}

type ReorderFormQuestionsPayload {
  form: Form
  clientMutationId: String
}

input SaveCaseInput {
  id: String
  workflow: ID!
  meta: JSONString
  parentWorkItem: ID
  form: ID
  clientMutationId: String
}

type SaveCasePayload {
  case: Case
  clientMutationId: String
}

input SaveChoiceQuestionInput {
  slug: String!
  label: String!
  infoText: String
  isRequired: QuestionJexl
  isHidden: QuestionJexl
  meta: JSONString
  isArchived: Boolean
  options: [ID]!
  clientMutationId: String
}

type SaveChoiceQuestionPayload {
  question: Question
  clientMutationId: String
}

input SaveCompleteTaskFormTaskInput {
  slug: String!
  name: String!
  description: String
  meta: JSONString
  addressGroups: GroupJexl
  controlGroups: GroupJexl
  isArchived: Boolean
  leadTime: Int
  isMultipleInstance: Boolean
  form: ID!
  clientMutationId: String
}

type SaveCompleteTaskFormTaskPayload {
  task: Task
  clientMutationId: String
}

input SaveCompleteWorkflowFormTaskInput {
  slug: String!
  name: String!
  description: String
  meta: JSONString
  addressGroups: GroupJexl
  controlGroups: GroupJexl
  isArchived: Boolean
  leadTime: Int
  isMultipleInstance: Boolean
  clientMutationId: String
}

type SaveCompleteWorkflowFormTaskPayload {
  task: Task
  clientMutationId: String
}

input SaveDateQuestionInput {
  slug: String!
  label: String!
  infoText: String
  isRequired: QuestionJexl
  isHidden: QuestionJexl
  meta: JSONString
  isArchived: Boolean
  clientMutationId: String
}

type SaveDateQuestionPayload {
  question: Question
  clientMutationId: String
}

input SaveDocumentDateAnswerInput {
  question: ID!
  document: ID!
  meta: JSONString
  value: Date
  clientMutationId: String
}

type SaveDocumentDateAnswerPayload {
  answer: Answer
  clientMutationId: String
}

input SaveDocumentFileAnswerInput {
  question: ID!
  document: ID!
  meta: JSONString
  value: String
  valueId: ID
  clientMutationId: String
}

type SaveDocumentFileAnswerPayload {
  answer: Answer
  clientMutationId: String
}

input SaveDocumentFloatAnswerInput {
  question: ID!
  document: ID!
  meta: JSONString
  value: Float
  clientMutationId: String
}

type SaveDocumentFloatAnswerPayload {
  answer: Answer
  clientMutationId: String
}

input SaveDocumentInput {
  id: String
  form: ID!
  meta: JSONString
  clientMutationId: String
}

input SaveDocumentIntegerAnswerInput {
  question: ID!
  document: ID!
  meta: JSONString
  value: Int
  clientMutationId: String
}

type SaveDocumentIntegerAnswerPayload {
  answer: Answer
  clientMutationId: String
}

input SaveDocumentListAnswerInput {
  question: ID!
  document: ID!
  meta: JSONString
  value: [String]
  clientMutationId: String
}

type SaveDocumentListAnswerPayload {
  answer: Answer
  clientMutationId: String
}

type SaveDocumentPayload {
  document: Document
  clientMutationId: String
}

input SaveDocumentStringAnswerInput {
  question: ID!
  document: ID!
  meta: JSONString
  value: String
  clientMutationId: String
}

type SaveDocumentStringAnswerPayload {
  answer: Answer
  clientMutationId: String
}

input SaveDocumentTableAnswerInput {
  question: ID!
  document: ID!
  meta: JSONString
  value: [ID]
  clientMutationId: String
}

type SaveDocumentTableAnswerPayload {
  answer: Answer
  clientMutationId: String
}

input SaveDynamicChoiceQuestionInput {
  slug: String!
  label: String!
  infoText: String
  isRequired: QuestionJexl
  isHidden: QuestionJexl
  meta: JSONString
  isArchived: Boolean
  dataSource: String!
  clientMutationId: String
}

type SaveDynamicChoiceQuestionPayload {
  question: Question
  clientMutationId: String
}

input SaveDynamicMultipleChoiceQuestionInput {
  slug: String!
  label: String!
  infoText: String
  isRequired: QuestionJexl
  isHidden: QuestionJexl
  meta: JSONString
  isArchived: Boolean
  dataSource: String!
  clientMutationId: String
}

type SaveDynamicMultipleChoiceQuestionPayload {
  question: Question
  clientMutationId: String
}

input SaveFileQuestionInput {
  slug: String!
  label: String!
  infoText: String
  isRequired: QuestionJexl
  isHidden: QuestionJexl
  meta: JSONString
  isArchived: Boolean
  clientMutationId: String
}

type SaveFileQuestionPayload {
  question: Question
  clientMutationId: String
}

input SaveFloatQuestionInput {
  slug: String!
  label: String!
  infoText: String
  isRequired: QuestionJexl
  isHidden: QuestionJexl
  meta: JSONString
  isArchived: Boolean
  minValue: Float
  maxValue: Float
  placeholder: String
  clientMutationId: String
}

type SaveFloatQuestionPayload {
  question: Question
  clientMutationId: String
}

input SaveFormInput {
  slug: String!
  name: String!
  description: String
  meta: JSONString
  isArchived: Boolean
  isPublished: Boolean
  clientMutationId: String
}

type SaveFormPayload {
  form: Form
  clientMutationId: String
}

input SaveFormQuestionInput {
  slug: String!
  label: String!
  infoText: String
  isRequired: QuestionJexl
  isHidden: QuestionJexl
  meta: JSONString
  isArchived: Boolean
  subForm: ID!
  clientMutationId: String
}

type SaveFormQuestionPayload {
  question: Question
  clientMutationId: String
}

input SaveIntegerQuestionInput {
  slug: String!
  label: String!
  infoText: String
  isRequired: QuestionJexl
  isHidden: QuestionJexl
  meta: JSONString
  isArchived: Boolean
  minValue: Int
  maxValue: Int
  placeholder: String
  clientMutationId: String
}

type SaveIntegerQuestionPayload {
  question: Question
  clientMutationId: String
}

input SaveMultipleChoiceQuestionInput {
  slug: String!
  label: String!
  infoText: String
  isRequired: QuestionJexl
  isHidden: QuestionJexl
  meta: JSONString
  isArchived: Boolean
  options: [ID]!
  clientMutationId: String
}

type SaveMultipleChoiceQuestionPayload {
  question: Question
  clientMutationId: String
}

input SaveOptionInput {
  slug: String!
  label: String!
  isArchived: Boolean
  meta: JSONString
  clientMutationId: String
}

type SaveOptionPayload {
  option: Option
  clientMutationId: String
}

input SaveSimpleTaskInput {
  slug: String!
  name: String!
  description: String
  meta: JSONString
  addressGroups: GroupJexl
  controlGroups: GroupJexl
  isArchived: Boolean
  leadTime: Int
  isMultipleInstance: Boolean
  clientMutationId: String
}

type SaveSimpleTaskPayload {
  task: Task
  clientMutationId: String
}

input SaveStaticQuestionInput {
  label: String!
  slug: String!
  infoText: String
  isHidden: QuestionJexl
  meta: JSONString
  isArchived: Boolean
  staticContent: String
  clientMutationId: String
}

type SaveStaticQuestionPayload {
  question: Question
  clientMutationId: String
}

input SaveTableQuestionInput {
  slug: String!
  label: String!
  infoText: String
  isRequired: QuestionJexl
  isHidden: QuestionJexl
  meta: JSONString
  isArchived: Boolean
  rowForm: ID!
  clientMutationId: String
}

type SaveTableQuestionPayload {
  question: Question
  clientMutationId: String
}

input SaveTextQuestionInput {
  slug: String!
  label: String!
  infoText: String
  isRequired: QuestionJexl
  isHidden: QuestionJexl
  meta: JSONString
  isArchived: Boolean
  minLength: Int
  maxLength: Int
  placeholder: String
  formatValidators: [String]
  clientMutationId: String
}

type SaveTextQuestionPayload {
  question: Question
  clientMutationId: String
}

input SaveTextareaQuestionInput {
  slug: String!
  label: String!
  infoText: String
  isRequired: QuestionJexl
  isHidden: QuestionJexl
  meta: JSONString
  isArchived: Boolean
  minLength: Int
  maxLength: Int
  placeholder: String
  formatValidators: [String]
  clientMutationId: String
}

type SaveTextareaQuestionPayload {
  question: Question
  clientMutationId: String
}

input SaveWorkItemInput {
  workItem: ID!
  name: String
  description: String
  assignedUsers: [String]
  deadline: DateTime
  meta: JSONString
  clientMutationId: String
}

type SaveWorkItemPayload {
  workItem: WorkItem
  clientMutationId: String
}

input SaveWorkflowInput {
  slug: String!
  name: String!
  description: String
  meta: JSONString
  startTasks: [ID]!
  allowAllForms: Boolean
  allowForms: [ID]
  isArchived: Boolean
  isPublished: Boolean
  clientMutationId: String
}

type SaveWorkflowPayload {
  workflow: Workflow
  clientMutationId: String
}

input SearchAnswersFilterType {
  questions: [ID]
  value: GenericScalar!
  lookup: SearchLookupMode
}

enum SearchLookupMode {
  STARTSWITH
  CONTAINS
  TEXT
}

type SimpleTask implements Task, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  slug: String!
  name: String!
  description: String
  type: TaskType!
  meta: GenericScalar!
  addressGroups: GroupJexl
  controlGroups: GroupJexl
  isArchived: Boolean!
  leadTime: Int
  isMultipleInstance: Boolean!
  id: ID!
}

input SkipWorkItemInput {
  id: ID!
  clientMutationId: String
}

type SkipWorkItemPayload {
  workItem: WorkItem
  clientMutationId: String
}

enum SortableAnswerAttributes {
  CREATED_AT
  MODIFIED_AT
  CREATED_BY_USER
  CREATED_BY_GROUP
  QUESTION
  VALUE
  DOCUMENT
  DATE
  FILE
}

enum SortableCaseAttributes {
  ALLOW_ALL_FORMS
  CREATED_BY_GROUP
  CREATED_BY_USER
  DESCRIPTION
  IS_ARCHIVED
  IS_PUBLISHED
  NAME
  STATUS
  SLUG
}

enum SortableDocumentAttributes {
  CREATED_AT
  MODIFIED_AT
  CREATED_BY_USER
  CREATED_BY_GROUP
  FORM
  SOURCE
}

enum SortableFormAttributes {
  CREATED_AT
  MODIFIED_AT
  CREATED_BY_USER
  CREATED_BY_GROUP
  SLUG
  NAME
  DESCRIPTION
  IS_PUBLISHED
  IS_ARCHIVED
}

enum SortableQuestionAttributes {
  CREATED_AT
  MODIFIED_AT
  CREATED_BY_USER
  CREATED_BY_GROUP
  SLUG
  LABEL
  TYPE
  IS_REQUIRED
  IS_HIDDEN
  IS_ARCHIVED
  PLACEHOLDER
  INFO_TEXT
}

enum SortableTaskAttributes {
  ALLOW_ALL_FORMS
  LEAD_TIME
  TYPE
  CREATED_BY_GROUP
  CREATED_BY_USER
  DESCRIPTION
  IS_ARCHIVED
  IS_PUBLISHED
  NAME
  SLUG
}

enum SortableWorkItemAttributes {
  ALLOW_ALL_FORMS
  CREATED_BY_GROUP
  CREATED_BY_USER
  DESCRIPTION
  CREATED_AT
  MODIFIED_AT
  CLOSED_AT
  IS_ARCHIVED
  IS_PUBLISHED
  NAME
  DEADLINE
  STATUS
  SLUG
}

enum SortableWorkflowAttributes {
  ALLOW_ALL_FORMS
  CREATED_BY_GROUP
  CREATED_BY_USER
  DESCRIPTION
  IS_ARCHIVED
  IS_PUBLISHED
  NAME
  SLUG
}

input StartCaseInput {
  workflow: ID!
  meta: JSONString
  parentWorkItem: ID
  form: ID
  clientMutationId: String
}

type StartCasePayload {
  case: Case
  clientMutationId: String
}

type StaticQuestion implements Question, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  slug: String!
  label: String!
  isRequired: QuestionJexl!
  isHidden: QuestionJexl!
  isArchived: Boolean!
  infoText: String
  meta: GenericScalar!
  source: Question
  forms(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], orderBy: [FormOrdering], slug: String, name: String, description: String, isPublished: Boolean, isArchived: Boolean, createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, search: String, slugs: [String]): FormConnection
  staticContent: String
  dataSource: String
  id: ID!
}

enum Status {
  READY
  COMPLETED
  CANCELED
  SKIPPED
}

type StringAnswer implements Answer, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  id: ID!
  question: Question!
  value: String
  meta: GenericScalar!
}

type TableAnswer implements Answer, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  id: ID!
  question: Question!
  value: [Document]
  meta: GenericScalar!
  document: Document!
}

type TableQuestion implements Question, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  slug: String!
  label: String!
  isRequired: QuestionJexl!
  isHidden: QuestionJexl!
  isArchived: Boolean!
  infoText: String
  meta: GenericScalar!
  source: Question
  forms(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], orderBy: [FormOrdering], slug: String, name: String, description: String, isPublished: Boolean, isArchived: Boolean, createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, search: String, slugs: [String]): FormConnection
  rowForm: Form
  id: ID!
}

interface Task {
  id: ID!
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  slug: String!
  name: String!
  description: String
  isArchived: Boolean!
  addressGroups: GroupJexl
  controlGroups: GroupJexl
  meta: GenericScalar!
  isMultipleInstance: Boolean!
}

type TaskConnection {
  pageInfo: PageInfo!
  edges: [TaskEdge]!
  totalCount: Int
}

type TaskEdge {
  node: Task
  cursor: String!
}

input TaskFilterSetType {
  slug: String
  name: String
  description: String
  type: Type
  isArchived: Boolean
  createdByUser: String
  createdByGroup: String
  createdBefore: DateTime
  createdAfter: DateTime
  metaHasKey: String
  metaValue: [JSONValueFilterType]
  search: String
  orderBy: [TaskOrdering]
  invert: Boolean
}

input TaskOrderSetType {
  meta: String
  attribute: SortableTaskAttributes
  direction: AscDesc
}

enum TaskOrdering {
  NAME_ASC
  NAME_DESC
  DESCRIPTION_ASC
  DESCRIPTION_DESC
  TYPE_ASC
  TYPE_DESC
  CREATED_AT_ASC
  CREATED_AT_DESC
  MODIFIED_AT_ASC
  MODIFIED_AT_DESC
  CREATED_BY_USER_ASC
  CREATED_BY_USER_DESC
  CREATED_BY_GROUP_ASC
  CREATED_BY_GROUP_DESC
  META_TEST_KEY_ASC
  META_TEST_KEY_DESC
  META_FOOBAR_ASC
  META_FOOBAR_DESC
}

enum TaskType {
  SIMPLE
  COMPLETE_WORKFLOW_FORM
  COMPLETE_TASK_FORM
}

enum TaskTypeArgument {
  SIMPLE
  COMPLETE_WORKFLOW_FORM
  COMPLETE_TASK_FORM
}

type TextQuestion implements Question, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  slug: String!
  label: String!
  isRequired: QuestionJexl!
  isHidden: QuestionJexl!
  isArchived: Boolean!
  placeholder: String
  infoText: String
  meta: GenericScalar!
  source: Question
  formatValidators(before: String, after: String, first: Int, last: Int): FormatValidatorConnection
  forms(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], orderBy: [FormOrdering], slug: String, name: String, description: String, isPublished: Boolean, isArchived: Boolean, createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, search: String, slugs: [String]): FormConnection
  id: ID!
  minLength: Int
  maxLength: Int
}

type TextareaQuestion implements Question, Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  slug: String!
  label: String!
  isRequired: QuestionJexl!
  isHidden: QuestionJexl!
  isArchived: Boolean!
  placeholder: String
  infoText: String
  meta: GenericScalar!
  source: Question
  formatValidators(before: String, after: String, first: Int, last: Int): FormatValidatorConnection
  forms(before: String, after: String, first: Int, last: Int, metaValue: [JSONValueFilterType], orderBy: [FormOrdering], slug: String, name: String, description: String, isPublished: Boolean, isArchived: Boolean, createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime, metaHasKey: String, search: String, slugs: [String]): FormConnection
  id: ID!
  minLength: Int
  maxLength: Int
}

enum Type {
  SIMPLE
  COMPLETE_WORKFLOW_FORM
  COMPLETE_TASK_FORM
}

scalar UUID

type ValidationEntry {
  slug: String!
  errorMsg: String!
}

type ValidationResult {
  id: ID
  isValid: Boolean
  errors: [ValidationEntry]
}

type WorkItem implements Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  id: ID!
  name: String!
  description: String
  closedAt: DateTime
  closedByUser: String
  closedByGroup: String
  deadline: DateTime
  task: Task!
  status: WorkItemStatus!
  meta: GenericScalar
  addressedGroups: [String]!
  controllingGroups: [String]!
  assignedUsers: [String]!
  case: Case!
  childCase: Case
  document: Document
}

type WorkItemConnection {
  pageInfo: PageInfo!
  edges: [WorkItemEdge]!
  totalCount: Int
}

type WorkItemEdge {
  node: WorkItem
  cursor: String!
}

input WorkItemFilterSetType {
  status: Status
  name: String
  task: ID
  case: ID
  createdAt: DateTime
  closedAt: DateTime
  modifiedAt: DateTime
  createdByUser: String
  createdByGroup: String
  createdBefore: DateTime
  createdAfter: DateTime
  metaHasKey: String
  metaValue: [JSONValueFilterType]
  orderBy: [WorkItemOrdering]
  addressedGroups: [String]
  controllingGroups: [String]
  assignedUsers: [String]
  documentHasAnswer: [HasAnswerFilterType]
  caseDocumentHasAnswer: [HasAnswerFilterType]
  caseMetaValue: [JSONValueFilterType]
  invert: Boolean
}

input WorkItemOrderSetType {
  meta: String
  caseMeta: String
  attribute: SortableWorkItemAttributes
  documentAnswer: String
  caseDocumentAnswer: String
  direction: AscDesc
}

enum WorkItemOrdering {
  STATUS_ASC
  STATUS_DESC
  DEADLINE_ASC
  DEADLINE_DESC
  CREATED_AT_ASC
  CREATED_AT_DESC
  MODIFIED_AT_ASC
  MODIFIED_AT_DESC
  CREATED_BY_USER_ASC
  CREATED_BY_USER_DESC
  CREATED_BY_GROUP_ASC
  CREATED_BY_GROUP_DESC
  META_TEST_KEY_ASC
  META_TEST_KEY_DESC
  META_FOOBAR_ASC
  META_FOOBAR_DESC
}

enum WorkItemStatus {
  READY
  COMPLETED
  CANCELED
  SKIPPED
}

enum WorkItemStatusArgument {
  READY
  COMPLETED
  CANCELED
  SKIPPED
}

type Workflow implements Node {
  createdAt: DateTime!
  modifiedAt: DateTime!
  createdByUser: String
  createdByGroup: String
  slug: String!
  name: String!
  description: String
  meta: GenericScalar
  isPublished: Boolean!
  isArchived: Boolean!
  startTasks: [Task]!
  allowAllForms: Boolean!
  allowForms(before: String, after: String, first: Int, last: Int): FormConnection!
  id: ID!
  tasks: [Task]!
  flows(before: String, after: String, first: Int, last: Int, task: ID, filter: [FlowFilterSetType], createdByUser: String, createdByGroup: String, createdBefore: DateTime, createdAfter: DateTime): FlowConnection
}

type WorkflowConnection {
  pageInfo: PageInfo!
  edges: [WorkflowEdge]!
  totalCount: Int
}

type WorkflowEdge {
  node: Workflow
  cursor: String!
}

input WorkflowFilterSetType {
  slug: String
  name: String
  description: String
  isPublished: Boolean
  isArchived: Boolean
  createdByUser: String
  createdByGroup: String
  createdBefore: DateTime
  createdAfter: DateTime
  metaHasKey: String
  metaValue: [JSONValueFilterType]
  search: String
  orderBy: [WorkflowOrdering]
  invert: Boolean
}

input WorkflowOrderSetType {
  meta: String
  attribute: SortableWorkflowAttributes
  direction: AscDesc
}

enum WorkflowOrdering {
  NAME_ASC
  NAME_DESC
  DESCRIPTION_ASC
  DESCRIPTION_DESC
  CREATED_AT_ASC
  CREATED_AT_DESC
  MODIFIED_AT_ASC
  MODIFIED_AT_DESC
  CREATED_BY_USER_ASC
  CREATED_BY_USER_DESC
  CREATED_BY_GROUP_ASC
  CREATED_BY_GROUP_DESC
  META_TEST_KEY_ASC
  META_TEST_KEY_DESC
  META_FOOBAR_ASC
  META_FOOBAR_DESC
}
"""



from django.apps import AppConfig
from django.conf import settings
from django.utils.module_loading import import_string


class DefaultConfig(AppConfig):
    name = "caluma.caluma_workflow"

    def ready(self):
        from .jexl import GroupJexl

        # to avoid recursive import error, load extension classes
        # only once the app is ready
        GroupJexl.dynamic_groups_classes = [
            import_string(cls) for cls in settings.DYNAMIC_GROUPS_CLASSES
        ]



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots["test_query_all_tasks[simple] 1"] = {
    "allTasks": {
        "edges": [
            {
                "node": {
                    "__typename": "SimpleTask",
                    "description": "State section rock event recent. Final activity hope star check record well. Radio with Mr letter eye.",
                    "meta": {},
                    "name": "Thomas Johnson",
                    "slug": "sound-air-mission",
                }
            }
        ],
        "totalCount": 1,
    }
}

snapshots["test_save_task[SaveSimpleTask] 1"] = {
    "saveSimpleTask": {
        "clientMutationId": "testid",
        "task": {
            "__typename": "SimpleTask",
            "meta": {},
            "name": "Thomas Johnson",
            "slug": "sound-air-mission",
        },
    }
}

snapshots["test_save_task[SaveCompleteWorkflowFormTask] 1"] = {
    "saveCompleteWorkflowFormTask": {
        "clientMutationId": "testid",
        "task": {
            "__typename": "CompleteWorkflowFormTask",
            "meta": {},
            "name": "Thomas Johnson",
            "slug": "sound-air-mission",
        },
    }
}

snapshots["test_save_comlete_task_form_task 1"] = {
    "saveCompleteTaskFormTask": {
        "clientMutationId": "testid",
        "task": {
            "__typename": "CompleteTaskFormTask",
            "meta": {},
            "name": "Thomas Johnson",
            "slug": "sound-air-mission",
        },
    }
}



import pytest

from ...caluma_form import models as form_models
from ...caluma_form.schema import Answer, Document
from .. import models
from ..schema import Case, WorkItem
from ..visibilities import AddressedGroups


@pytest.mark.parametrize(
    "work_item__addressed_groups,size", [(["unknown"], 0), (["admin", "other"], 1)]
)
def test_assigned_groups_work_item_visibility(db, admin_info, size, work_item):

    queryset = AddressedGroups().filter_queryset(
        WorkItem, models.WorkItem.objects, admin_info
    )
    assert queryset.count() == size


@pytest.mark.parametrize(
    "work_item__addressed_groups,size", [(["unknown"], 0), (["admin", "other"], 1)]
)
def test_assigned_groups_case_visibility(db, admin_info, size, work_item):

    queryset = AddressedGroups().filter_queryset(Case, models.Case.objects, admin_info)
    assert queryset.count() == size


@pytest.mark.parametrize(
    "work_item__addressed_groups,size", [(["unknown"], 0), (["admin", "other"], 1)]
)
def test_assigned_groups_document_visibility(db, admin_info, size, work_item):

    queryset = AddressedGroups().filter_queryset(
        Document, form_models.Document.objects, admin_info
    )
    assert queryset.count() == size


@pytest.mark.parametrize(
    "work_item__addressed_groups,size", [(["unknown"], 0), (["admin", "other"], 1)]
)
def test_assigned_groups_answer_visibility(db, admin_info, size, work_item, answer):

    queryset = AddressedGroups().filter_queryset(
        Answer, form_models.Answer.objects, admin_info
    )
    assert queryset.count() == size



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots["test_document_as_of 1"] = {
    "documentAsOf": {
        "documentId": "890ca108-d93d-4725-9066-7d0bddad8230",
        "historicalAnswers": {
            "edges": [
                {
                    "node": {
                        "__typename": "HistoricalStringAnswer",
                        "historyUserId": "admin",
                        "value": "first admin - revision 1",
                    }
                }
            ]
        },
        "meta": {},
    }
}

snapshots["test_document_as_of 2"] = {
    "documentAsOf": {
        "documentId": "890ca108-d93d-4725-9066-7d0bddad8230",
        "historicalAnswers": {
            "edges": [
                {
                    "node": {
                        "__typename": "HistoricalStringAnswer",
                        "historyUserId": "AnonymousUser",
                        "value": "first anon - revision 3",
                    }
                }
            ]
        },
        "meta": {},
    }
}

snapshots["test_document_as_of 3"] = {
    "documentAsOf": {
        "documentId": "890ca108-d93d-4725-9066-7d0bddad8230",
        "historicalAnswers": {
            "edges": [
                {
                    "node": {
                        "__typename": "HistoricalStringAnswer",
                        "historyUserId": "AnonymousUser",
                        "value": "second anon - revision 4",
                    }
                }
            ]
        },
        "meta": {},
    }
}

snapshots["test_historical_table_answer 1"] = {
    "d1": {
        "historicalAnswers": {
            "edges": [
                {
                    "node": {
                        "__typename": "HistoricalTableAnswer",
                        "value": [
                            {
                                "historicalAnswers": {
                                    "edges": [
                                        {
                                            "node": {
                                                "historyType": "+",
                                                "value": "first row value",
                                            }
                                        }
                                    ]
                                }
                            },
                            {
                                "historicalAnswers": {
                                    "edges": [
                                        {
                                            "node": {
                                                "historyType": "+",
                                                "value": "second row value",
                                            }
                                        }
                                    ]
                                }
                            },
                        ],
                    }
                }
            ]
        }
    },
    "d2": {
        "historicalAnswers": {
            "edges": [
                {
                    "node": {
                        "__typename": "HistoricalTableAnswer",
                        "value": [
                            {
                                "historicalAnswers": {
                                    "edges": [
                                        {
                                            "node": {
                                                "historyType": "+",
                                                "value": "first row value",
                                            }
                                        }
                                    ]
                                }
                            },
                            {
                                "historicalAnswers": {
                                    "edges": [
                                        {
                                            "node": {
                                                "historyType": "-",
                                                "value": "second row value",
                                            }
                                        }
                                    ]
                                }
                            },
                        ],
                    }
                }
            ]
        }
    },
}



# -*- coding: utf-8 -*-
# Generated by Django 1.11.21 on 2019-06-24 11:14
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0018_remove_answer_value_document_data_migration")]

    operations = [migrations.RemoveField(model_name="answer", name="value_document")]



from functools import wraps


def register_dynamic_group(dynamic_group_name):
    def wrapper(fn, *arg, **kwarg):
        fn._dynamic_group_name = dynamic_group_name

        @wraps(fn)
        def wrapped(self, *args, **kwargs):
            return fn(self, *args, **kwargs)

        return wrapped

    return wrapper


class BaseDynamicGroups:
    def resolve(self, dynamic_group_name):
        for methodname in dir(self):
            method = getattr(self, methodname)
            if (
                hasattr(method, "_dynamic_group_name")
                and method._dynamic_group_name == dynamic_group_name
            ):
                return method



from django.core.exceptions import ValidationError
from django.utils import timezone

from caluma.caluma_core.events import send_event

from ..caluma_form.models import Document
from . import events, models, utils, validators
from .jexl import FlowJexl


class StartCaseLogic:
    """
    Shared domain logic for starting cases.

    Used in the `saveCase` mutation and in the `start_case` API. The logic
    for case creation is split in three parts (`validate_for_start`,
    `pre_start` and `post_start`) so that in between the appropriate create
    method can be called (`super().create(...)` for the serializer and
    `Case.objects.create(...) for the python API`).
    """

    @staticmethod
    def validate_for_start(data):
        form = data.get("form")
        workflow = data.get("workflow")

        if form:
            if (
                not workflow.allow_all_forms
                and not workflow.allow_forms.filter(pk=form.pk).exists()
            ):
                raise ValidationError(
                    f"Workflow {workflow.pk} does not allow to start case with form {form.pk}"
                )

        return data

    @staticmethod
    def pre_start(validated_data, user):
        parent_work_item = validated_data.get("parent_work_item")
        validated_data["status"] = models.Case.STATUS_RUNNING

        form = validated_data.pop("form", None)
        if form:
            validated_data["document"] = Document.objects.create(
                form=form, created_by_user=user.username, created_by_group=user.group
            )

        if parent_work_item:
            case = parent_work_item.case
            while hasattr(case, "parent_work_item"):
                case = case.parent_work_item.case
            validated_data["family"] = case

        return validated_data

    @staticmethod
    def post_start(case, user, parent_work_item):
        # Django doesn't save reverse one-to-one relationships automatically:
        # https://code.djangoproject.com/ticket/18638
        if parent_work_item:
            parent_work_item.child_case = case
            parent_work_item.save()

        workflow = case.workflow
        tasks = workflow.start_tasks.all()

        work_items = utils.bulk_create_work_items(tasks, case, user)

        send_event(events.created_case, sender="case_post_create", case=case)
        for work_item in work_items:  # pragma: no cover
            send_event(
                events.created_work_item, sender="case_post_create", work_item=work_item
            )

        return case


class CompleteWorkItemLogic:
    @staticmethod
    def _can_continue(work_item, task):
        # If a "multiple instance" task has running siblings, the task is not completed
        if task.is_multiple_instance:
            return not work_item.case.work_items.filter(
                task=task, status=models.WorkItem.STATUS_READY
            ).exists()
        return work_item.case.work_items.filter(
            task=task,
            status__in=(
                models.WorkItem.STATUS_COMPLETED,
                models.WorkItem.STATUS_SKIPPED,
            ),
        ).exists()

    @staticmethod
    def validate_for_complete(work_item, user):
        validators.WorkItemValidator().validate(
            status=work_item.status,
            child_case=work_item.child_case,
            task=work_item.task,
            case=work_item.case,
            document=work_item.document,
            user=user,
        )

    @staticmethod
    def pre_complete(validated_data, user):
        validated_data["status"] = models.WorkItem.STATUS_COMPLETED
        validated_data["closed_at"] = timezone.now()
        validated_data["closed_by_user"] = user.username
        validated_data["closed_by_group"] = user.group

        return validated_data

    @staticmethod
    def post_complete(work_item, user):
        case = work_item.case

        if not CompleteWorkItemLogic._can_continue(work_item, work_item.task):
            return work_item

        flow = models.Flow.objects.filter(
            task_flows__task=work_item.task_id,
            task_flows__workflow=work_item.case.workflow.pk,
        ).first()

        if flow:
            sibling_tasks = models.Task.objects.filter(task_flows__flow=flow)

            all_siblings_complete = all(
                CompleteWorkItemLogic._can_continue(work_item, task)
                for task in sibling_tasks
            )

            if all_siblings_complete:
                jexl = FlowJexl()
                result = jexl.evaluate(flow.next)
                if not isinstance(result, list):
                    result = [result]

                tasks = models.Task.objects.filter(pk__in=result)

                created_work_items = utils.bulk_create_work_items(
                    tasks, case, user, work_item
                )

                for created_work_item in created_work_items:  # pragma: no cover
                    send_event(
                        events.created_work_item,
                        sender="post_complete_work_item",
                        work_item=created_work_item,
                    )
        else:
            has_ready_work_items = work_item.case.work_items.filter(
                status=models.WorkItem.STATUS_READY
            ).exists()

            if not has_ready_work_items:
                # no more tasks, mark case as complete
                case.status = models.Case.STATUS_COMPLETED
                case.closed_at = timezone.now()
                case.closed_by_user = user.username
                case.closed_by_group = user.group
                case.save()
                send_event(
                    events.completed_case, sender="post_complete_work_item", case=case
                )

        send_event(
            events.completed_work_item,
            sender="post_complete_work_item",
            work_item=work_item,
        )

        return work_item


class SkipWorkItemLogic:
    @staticmethod
    def validate_for_skip(work_item):
        if work_item.status != models.WorkItem.STATUS_READY:
            raise ValidationError("Only READY work items can be skipped")

    @staticmethod
    def pre_skip(validated_data, user):
        validated_data = CompleteWorkItemLogic.pre_complete(validated_data, user)

        validated_data["status"] = models.WorkItem.STATUS_SKIPPED

        return validated_data

    @staticmethod
    def post_skip(work_item, user):
        work_item = CompleteWorkItemLogic.post_complete(work_item, user)

        send_event(
            events.skipped_work_item, sender="post_skip_work_item", work_item=work_item
        )

        return work_item


class CancelCaseLogic:
    """
    Shared domain logic for cancelling cases.

    Used in the `cancelCase` mutation and in the `cancel_case` API. The logic
    for case cancellation is split in three parts (`validate_for_cancel`,
    `pre_cancel` and `post_cancel`) so that in between the appropriate update
    method can be called (`super().update(...)` for the serializer and
    `Case.objects.update(...) for the python API`).
    """

    @staticmethod
    def validate_for_cancel(case):
        if case.status != models.Case.STATUS_RUNNING:
            raise ValidationError("Only running cases can be canceled.")

    @staticmethod
    def pre_cancel(validated_data, user):
        validated_data["status"] = models.Case.STATUS_CANCELED
        validated_data["closed_at"] = timezone.now()
        validated_data["closed_by_user"] = user.username
        validated_data["closed_by_group"] = user.group

        return validated_data

    @staticmethod
    def post_cancel(case, user):
        work_items = case.work_items.exclude(
            status__in=[
                models.WorkItem.STATUS_COMPLETED,
                models.WorkItem.STATUS_CANCELED,
            ]
        )

        for work_item in work_items:
            work_item.status = models.WorkItem.STATUS_CANCELED
            work_item.closed_at = timezone.now()
            work_item.closed_by_user = user.username
            work_item.closed_by_group = user.group
            work_item.save()

        # send events in separate loop in order to be sure all operations are finished
        for work_item in work_items:
            send_event(
                events.cancelled_work_item,
                sender="post_cancel_case",
                work_item=work_item,
            )

        send_event(events.cancelled_case, sender="post_cancel_case", case=case)

        return case



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot

snapshots = Snapshot()

snapshots["test_query_all_workflows 1"] = {
    "allWorkflows": {
        "edges": [
            {
                "node": {
                    "allowAllForms": False,
                    "allowForms": {"edges": [{"node": {"slug": "sound-air-mission"}}]},
                    "description": """Line whatever team suggest traditional boy. Drop argue move. Anyone remember prove.
Kid avoid player relationship to range whose. Draw free property consider.""",
                    "flows": {
                        "edges": [
                            {
                                "node": {
                                    "next": "mrs-positive",
                                    "tasks": [{"slug": "few-list-tax"}],
                                }
                            }
                        ],
                        "totalCount": 1,
                    },
                    "meta": {},
                    "name": "Brian Williams",
                    "slug": "effort-meet",
                    "startTasks": [{"slug": "few-list-tax"}],
                    "tasks": [{"slug": "few-list-tax"}],
                }
            }
        ],
        "totalCount": 1,
    }
}

snapshots["test_save_workflow 1"] = {
    "saveWorkflow": {
        "clientMutationId": "testid",
        "workflow": {
            "allowAllForms": False,
            "allowForms": {
                "edges": [{"node": {"slug": "sound-air-mission"}}],
                "totalCount": 1,
            },
            "meta": {},
            "name": "Brian Williams",
            "slug": "effort-meet",
        },
    }
}

snapshots['test_add_workflow_flow[task-slug-"task-slug"|task-True] 1'] = {
    "addWorkflowFlow": {
        "clientMutationId": None,
        "workflow": {
            "flows": {
                "edges": [
                    {
                        "node": {
                            "createdByGroup": "admin",
                            "createdByUser": "admin",
                            "next": '"task-slug"|task',
                            "tasks": [{"slug": "task-slug"}],
                        }
                    }
                ],
                "totalCount": 1,
            },
            "tasks": [
                {"slug": "task-slug"},
                {"slug": "service-bank-arm"},
                {"slug": "relate-right-will"},
            ],
        },
    }
}



# Generated by Django 2.2.10 on 2020-03-13 09:23

import django.contrib.postgres.fields
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_workflow", "0020_case_families")]

    operations = [
        migrations.AddField(
            model_name="historicaltask",
            name="control_groups",
            field=models.TextField(
                blank=True,
                help_text="Group jexl returning what group(s) derived work items will be assigned to for controlling.",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="historicalworkitem",
            name="controlling_groups",
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=150),
                default=list,
                help_text="List of groups this work item is assigned to for controlling.",
                size=None,
            ),
        ),
        migrations.AddField(
            model_name="task",
            name="control_groups",
            field=models.TextField(
                blank=True,
                help_text="Group jexl returning what group(s) derived work items will be assigned to for controlling.",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="workitem",
            name="controlling_groups",
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=150),
                default=list,
                help_text="List of groups this work item is assigned to for controlling.",
                size=None,
            ),
        ),
    ]



from functools import partial
from itertools import chain

from pyjexl.analysis import ValidatingAnalyzer
from pyjexl.evaluator import Context

from ..caluma_core.jexl import JEXL, ExtractTransformSubjectAnalyzer


class GroupValidatingAnalyzer(ValidatingAnalyzer):
    def visit_Transform(self, transform):
        if transform.name == "groups" and not isinstance(transform.subject.value, list):
            yield f"{transform.subject.value} is not a valid list of groups."

        yield from super().visit_Transform(transform)


class TaskValidatingAnalyzer(ValidatingAnalyzer):
    def visit_Transform(self, transform):
        if transform.name == "tasks" and not isinstance(transform.subject.value, list):
            yield f"{transform.subject.value} is not a valid list of tasks."

        if transform.name == "task" and not isinstance(transform.subject.value, str):
            yield f"{transform.subject.value} is not a valid task slug."

        yield from super().visit_Transform(transform)


class GroupJexl(JEXL):
    """
    Class for evaluating GroupJexl.

    validation_context is expected to be the following:

    {
        "case": {
            "created_by_group": str,
        },
        "work_item": {
            "created_by_group": str,
        },
        "prev_work_item": {
            "controlling_groups": list of str,
        },
    }
    """

    def __init__(
        self,
        validation_context=None,
        task=None,
        case=None,
        work_item_created_by_user=None,
        prev_work_item=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        context_data = None

        if validation_context:
            context_data = {"info": validation_context}

        self.context = Context(context_data)
        self.add_transform(
            "groups",
            lambda names: list(
                set(
                    chain(
                        *[
                            self._get_dynamic_group(
                                name,
                                task,
                                case,
                                work_item_created_by_user,
                                prev_work_item,
                            )
                            for name in names
                        ]
                    )
                )
            ),
        )

    def validate(self, expression):
        return super().validate(expression, GroupValidatingAnalyzer)

    def evaluate(self, expression, context=None):
        value = super().evaluate(expression, context)
        if isinstance(value, list) or value is None:
            return value
        return [value]

    def _get_dynamic_group(
        self, name, task, case, work_item_created_by_user, prev_work_item
    ):
        for dynamic_groups_class in self.dynamic_groups_classes:
            method = dynamic_groups_class().resolve(name)
            if method:
                value = method(task, case, work_item_created_by_user, prev_work_item)

                if not isinstance(value, list):
                    value = [value]

                return value

        return [name]


class FlowJexl(JEXL):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.add_transform("task", lambda spec: spec)
        self.add_transform("tasks", lambda spec: spec)

    def validate(self, expression):
        return super().validate(expression, TaskValidatingAnalyzer)

    def extract_tasks(self, expr):
        yield from self.analyze(
            expr, partial(ExtractTransformSubjectAnalyzer, transforms=["task"])
        )

        # tasks transforms return a list of literals
        yield from (
            literal.value
            for literal in chain(
                *self.analyze(
                    expr, partial(ExtractTransformSubjectAnalyzer, transforms=["tasks"])
                )
            )
        )



from django.core.exceptions import ValidationError

from ..caluma_form.validators import DocumentValidator
from . import models


class WorkItemValidator:
    def _validate_task_simple(self, task, case, document, user):
        pass

    def _validate_task_complete_workflow_form(self, task, case, document, user):
        self._validate_task_complete_task_form(task, case, case.document, user)

    def _validate_task_complete_task_form(self, task, case, document, user):
        DocumentValidator().validate(document, user)

    def validate(self, *, status, child_case, case, task, document, user, **kwargs):
        if status != models.WorkItem.STATUS_READY:
            raise ValidationError("Only ready work items can be completed.")

        if child_case:
            if child_case.status == models.Case.STATUS_RUNNING:
                raise ValidationError(
                    "Work item can only be completed when child case is in a finish state."
                )

        getattr(self, f"_validate_task_{task.type}")(task, case, document, user)



import pytest
from django.utils import translation

from ...caluma_core.tests import extract_serializer_input_fields
from ..format_validators import BaseFormatValidator, base_format_validators
from ..models import Question
from ..serializers import SaveAnswerSerializer


class MyFormatValidator(BaseFormatValidator):
    slug = "test-validator"
    name = {"en": "test name english", "de": "test name deutsch"}
    regex = r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)"
    error_msg = {"en": "Not valid", "de": "Nicht valid"}


def test_fetch_format_validators(snapshot, schema_executor, settings):
    settings.FORMAT_VALIDATOR_CLASSES = [
        "caluma.caluma_form.tests.test_format_validators.MyFormatValidator"
    ]
    query = """
        query formatValidators {
          allFormatValidators {
            pageInfo {
              startCursor
              endCursor
            }
            totalCount
            edges {
              node {
                slug
                name
                regex
                errorMsg
              }
            }
          }
        }
    """

    result = schema_executor(query)
    assert not result.errors
    snapshot.assert_match(result.data)

    with translation.override("de"):
        result = schema_executor(query)
        assert not result.errors
        snapshot.assert_match(result.data)

    with translation.override("fr"):
        result = schema_executor(query)
        assert not result.errors
        snapshot.assert_match(result.data)

    with translation.override("nolang"):
        result = schema_executor(query)
        assert not result.errors
        snapshot.assert_match(result.data)


@pytest.mark.parametrize("question__type", [Question.TYPE_TEXT, Question.TYPE_TEXTAREA])
@pytest.mark.parametrize(
    "question__format_validators,answer__value,success",
    [
        (["email"], "some text", False),
        (["email"], "test@example.com", True),
        (["phone-number"], "some text", False),
        (["phone-number"], "+411234567890", True),
        (["email", "phone-number"], "+411234567890", False),
        (["email", "phone-number"], "test@example.com", False),
    ],
)
def test_base_format_validators(
    db, snapshot, question, answer, success, schema_executor
):
    query = """
        mutation SaveDocumentStringAnswer($input: SaveDocumentStringAnswerInput!) {
          saveDocumentStringAnswer(input: $input) {
            answer {
              ... on StringAnswer {
                stringValue: value
              }
            }
            clientMutationId
          }
        }
    """

    inp = {"input": extract_serializer_input_fields(SaveAnswerSerializer, answer)}

    result = schema_executor(query, variable_values=inp)

    assert not bool(result.errors) == success
    if success:
        snapshot.assert_match(result.data)
    if not success:
        error_msgs = [fv.error_msg["en"] for fv in base_format_validators]
        assert (
            str(result.errors[0].original_error.detail["non_field_errors"][0])
            in error_msgs
        )



from django.core.management.base import BaseCommand
from django.db import connection
from django.db.utils import ProgrammingError


class Command(BaseCommand):
    """Migrate db to prefixed apps."""

    help = "Migrate db to prefixed apps."

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.revert = False
        self.queries = []
        self.prefix = "caluma_"

    def add_arguments(self, parser):
        parser.add_argument(
            "-f", "--force", action="store_true", help="Do the migration"
        )
        parser.add_argument(
            "-r", "--revert", action="store_true", help="Revert the migration"
        )

    @property
    def _act_prefix(self):
        prefix = ""
        if self.revert:
            prefix = self.prefix
        return prefix

    def _migration_needed(self):
        query = f"""SELECT "app" FROM "django_migrations" WHERE "app" = '{self._act_prefix}form';"""
        with connection.cursor() as cursor:
            try:
                cursor.execute(query)
            except ProgrammingError as e:  # pragma: no cover
                # happens only on initial migration for new project
                if e.args[0].startswith('relation "django_migrations" does not exist'):
                    return False
                raise
            if cursor.fetchone():
                return True
        return False

    def _get_table_list(self):
        query = f"""SELECT tablename FROM pg_catalog.pg_tables WHERE tablename LIKE '{self._act_prefix}form_%' OR tablename LIKE '{self._act_prefix}workflow_%';"""
        with connection.cursor() as cursor:
            cursor.execute(query)
            return [row[0] for row in cursor.fetchall()]

    def _collect_content_type_queries(self):
        queries = [
            f"""UPDATE django_content_type SET app_label = CONCAT('{self.prefix}', app_label) WHERE app_label = 'form' OR app_label = 'workflow';"""
        ]
        if self.revert:
            queries = [
                f"""UPDATE django_content_type SET app_label = REPLACE(app_label, 'caluma_', '');"""
            ]
        return queries

    def _collect_django_migrations_queries(self):
        queries = [
            f"""UPDATE django_migrations SET app = CONCAT('{self.prefix}', app) WHERE app = 'form' OR app = 'workflow';"""
        ]
        if self.revert:
            queries = [
                f"""UPDATE django_migrations SET app = REPLACE(app, 'caluma_', '');"""
            ]
        return queries

    def _collect_prefix_tables_queries(self, tables: list):
        queries = []
        for table in tables:
            query = f"""ALTER TABLE {table} RENAME TO {self.prefix}{table};"""
            if self.revert:
                query = f"""ALTER TABLE {table} RENAME TO {table.replace(self.prefix, '')};"""
            queries.append(query)
        return queries

    def collect(self):
        if not self._migration_needed():
            return
        self.queries += self._collect_content_type_queries()
        self.queries += self._collect_django_migrations_queries()
        tables = self._get_table_list()
        self.queries += self._collect_prefix_tables_queries(tables)
        return self.queries

    def apply(self):
        with connection.cursor() as cursor:
            for query in self.queries:
                cursor.execute(query)

    def handle(self, *args, **options):
        if options["revert"]:
            self.revert = True

        self.collect()

        if not self.queries:
            self.stderr.write("Seems the DB is already in the desired state!")
            return

        self.stderr.write("\n".join(self.queries))

        if options["force"]:
            self.apply()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.18 on 2019-02-08 10:16
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("caluma_form", "0004_auto_20190207_1405")]

    operations = [
        migrations.AlterField(
            model_name="question",
            name="type",
            field=models.CharField(
                choices=[
                    ("multiple_choice", "multiple_choice"),
                    ("integer", "integer"),
                    ("float", "float"),
                    ("choice", "choice"),
                    ("textarea", "textarea"),
                    ("text", "text"),
                    ("table", "table"),
                ],
                max_length=15,
            ),
        )
    ]



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots["test_query_all_cases[running-1] 1"] = {
    "allCases": {"edges": [{"node": {"status": "RUNNING"}}], "totalCount": 1}
}

snapshots["test_query_all_cases[completed-0] 1"] = {
    "allCases": {"edges": [], "totalCount": 0}
}

snapshots['test_start_case[startCase-["group-name"]|groups-100] 1'] = {
    "startCase": {
        "case": {
            "document": {"form": {"slug": "sound-air-mission"}},
            "parentWorkItem": None,
            "status": "RUNNING",
            "workItems": {
                "edges": [
                    {
                        "node": {
                            "addressedGroups": ["group-name"],
                            "document": {"form": {"slug": "sound-air-mission"}},
                            "status": "READY",
                        }
                    }
                ]
            },
        },
        "clientMutationId": None,
    }
}

snapshots['test_start_case[startCase-["group-name"]|groups-None] 1'] = {
    "startCase": {
        "case": {
            "document": {"form": {"slug": "sound-air-mission"}},
            "parentWorkItem": None,
            "status": "RUNNING",
            "workItems": {
                "edges": [
                    {
                        "node": {
                            "addressedGroups": ["group-name"],
                            "document": {"form": {"slug": "sound-air-mission"}},
                            "status": "READY",
                        }
                    }
                ]
            },
        },
        "clientMutationId": None,
    }
}

snapshots["test_start_case[startCase-None-100] 1"] = {
    "startCase": {
        "case": {
            "document": {"form": {"slug": "sound-air-mission"}},
            "parentWorkItem": None,
            "status": "RUNNING",
            "workItems": {
                "edges": [
                    {
                        "node": {
                            "addressedGroups": [],
                            "document": {"form": {"slug": "sound-air-mission"}},
                            "status": "READY",
                        }
                    }
                ]
            },
        },
        "clientMutationId": None,
    }
}

snapshots["test_start_case[startCase-None-None] 1"] = {
    "startCase": {
        "case": {
            "document": {"form": {"slug": "sound-air-mission"}},
            "parentWorkItem": None,
            "status": "RUNNING",
            "workItems": {
                "edges": [
                    {
                        "node": {
                            "addressedGroups": [],
                            "document": {"form": {"slug": "sound-air-mission"}},
                            "status": "READY",
                        }
                    }
                ]
            },
        },
        "clientMutationId": None,
    }
}

snapshots['test_start_case[saveCase-["group-name"]|groups-100] 1'] = {
    "saveCase": {
        "case": {
            "document": {"form": {"slug": "sound-air-mission"}},
            "parentWorkItem": {"status": "READY"},
            "status": "RUNNING",
            "workItems": {
                "edges": [
                    {
                        "node": {
                            "addressedGroups": [],
                            "document": {"form": {"slug": "sound-air-mission"}},
                            "status": "READY",
                        }
                    }
                ]
            },
        },
        "clientMutationId": None,
    }
}

snapshots['test_start_case[saveCase-["group-name"]|groups-None] 1'] = {
    "saveCase": {
        "case": {
            "document": {"form": {"slug": "sound-air-mission"}},
            "parentWorkItem": {"status": "READY"},
            "status": "RUNNING",
            "workItems": {
                "edges": [
                    {
                        "node": {
                            "addressedGroups": [],
                            "document": {"form": {"slug": "sound-air-mission"}},
                            "status": "READY",
                        }
                    }
                ]
            },
        },
        "clientMutationId": None,
    }
}

snapshots["test_start_case[saveCase-None-100] 1"] = {
    "saveCase": {
        "case": {
            "document": {"form": {"slug": "sound-air-mission"}},
            "parentWorkItem": {"status": "READY"},
            "status": "RUNNING",
            "workItems": {
                "edges": [
                    {
                        "node": {
                            "addressedGroups": [],
                            "document": {"form": {"slug": "sound-air-mission"}},
                            "status": "READY",
                        }
                    }
                ]
            },
        },
        "clientMutationId": None,
    }
}

snapshots["test_start_case[saveCase-None-None] 1"] = {
    "saveCase": {
        "case": {
            "document": {"form": {"slug": "sound-air-mission"}},
            "parentWorkItem": {"status": "READY"},
            "status": "RUNNING",
            "workItems": {
                "edges": [
                    {
                        "node": {
                            "addressedGroups": [],
                            "document": {"form": {"slug": "sound-air-mission"}},
                            "status": "READY",
                        }
                    }
                ]
            },
        },
        "clientMutationId": None,
    }
}

snapshots["test_cancel_case[running-True-completed] 1"] = {
    "cancelCase": {
        "case": {
            "document": {"form": {"slug": "sound-air-mission"}},
            "status": "CANCELED",
            "workItems": {"edges": [{"node": {"status": "COMPLETED"}}]},
        },
        "clientMutationId": None,
    }
}

snapshots["test_cancel_case[running-True-ready] 1"] = {
    "cancelCase": {
        "case": {
            "document": {"form": {"slug": "sound-air-mission"}},
            "status": "CANCELED",
            "workItems": {"edges": [{"node": {"status": "CANCELED"}}]},
        },
        "clientMutationId": None,
    }
}

snapshots["test_order_by_question_answer_value[text-True-True] 1"] = {
    "allCases": {
        "edges": [
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [
                                {"node": {"stringValue": "a"}},
                                {"node": {"stringValue": "b2"}},
                            ],
                            "totalCount": 2,
                        }
                    }
                }
            },
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [
                                {"node": {"stringValue": "b"}},
                                {"node": {"stringValue": "c2"}},
                            ],
                            "totalCount": 2,
                        }
                    }
                }
            },
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [
                                {"node": {"stringValue": "c"}},
                                {"node": {"stringValue": "a2"}},
                            ],
                            "totalCount": 2,
                        }
                    }
                }
            },
        ],
        "totalCount": 3,
    }
}

snapshots["test_order_by_question_answer_value[text-True-False] 1"] = {
    "allCases": {
        "edges": [
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [
                                {"node": {"stringValue": "c"}},
                                {"node": {"stringValue": "a2"}},
                            ],
                            "totalCount": 2,
                        }
                    }
                }
            },
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [
                                {"node": {"stringValue": "b"}},
                                {"node": {"stringValue": "c2"}},
                            ],
                            "totalCount": 2,
                        }
                    }
                }
            },
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [
                                {"node": {"stringValue": "a"}},
                                {"node": {"stringValue": "b2"}},
                            ],
                            "totalCount": 2,
                        }
                    }
                }
            },
        ],
        "totalCount": 3,
    }
}

snapshots["test_order_by_question_answer_value[form-True-True] 1"] = {
    "allCases": {
        "edges": [
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [{"node": {"stringValue": "b"}}],
                            "totalCount": 1,
                        }
                    }
                }
            },
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [{"node": {"stringValue": "d"}}],
                            "totalCount": 1,
                        }
                    }
                }
            },
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [{"node": {"stringValue": "f"}}],
                            "totalCount": 1,
                        }
                    }
                }
            },
        ],
        "totalCount": 3,
    }
}

snapshots["test_order_by_question_answer_value[form-True-False] 1"] = {
    "allCases": {
        "edges": [
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [{"node": {"stringValue": "f"}}],
                            "totalCount": 1,
                        }
                    }
                }
            },
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [{"node": {"stringValue": "d"}}],
                            "totalCount": 1,
                        }
                    }
                }
            },
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [{"node": {"stringValue": "b"}}],
                            "totalCount": 1,
                        }
                    }
                }
            },
        ],
        "totalCount": 3,
    }
}

snapshots["test_order_by_question_answer_value[date-True-True] 1"] = {
    "allCases": {
        "edges": [
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [
                                {"node": {"dateValue": "2019-05-27"}},
                                {"node": {"dateValue": "2019-05-28"}},
                            ],
                            "totalCount": 2,
                        }
                    }
                }
            },
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [
                                {"node": {"dateValue": "2019-05-29"}},
                                {"node": {"dateValue": "2019-05-30"}},
                            ],
                            "totalCount": 2,
                        }
                    }
                }
            },
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [
                                {"node": {"dateValue": "2019-05-31"}},
                                {"node": {"dateValue": "2019-05-26"}},
                            ],
                            "totalCount": 2,
                        }
                    }
                }
            },
        ],
        "totalCount": 3,
    }
}

snapshots["test_order_by_question_answer_value[date-True-False] 1"] = {
    "allCases": {
        "edges": [
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [
                                {"node": {"dateValue": "2019-05-31"}},
                                {"node": {"dateValue": "2019-05-26"}},
                            ],
                            "totalCount": 2,
                        }
                    }
                }
            },
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [
                                {"node": {"dateValue": "2019-05-29"}},
                                {"node": {"dateValue": "2019-05-30"}},
                            ],
                            "totalCount": 2,
                        }
                    }
                }
            },
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [
                                {"node": {"dateValue": "2019-05-27"}},
                                {"node": {"dateValue": "2019-05-28"}},
                            ],
                            "totalCount": 2,
                        }
                    }
                }
            },
        ],
        "totalCount": 3,
    }
}

snapshots["test_order_by_question_answer_value[file-True-True] 1"] = {
    "allCases": {
        "edges": [
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [
                                {"node": {"fileValue": {"name": "b"}}},
                                {"node": {"fileValue": {"name": "e"}}},
                            ],
                            "totalCount": 2,
                        }
                    }
                }
            },
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [
                                {"node": {"fileValue": {"name": "d"}}},
                                {"node": {"fileValue": {"name": "c"}}},
                            ],
                            "totalCount": 2,
                        }
                    }
                }
            },
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [
                                {"node": {"fileValue": {"name": "f"}}},
                                {"node": {"fileValue": {"name": "a"}}},
                            ],
                            "totalCount": 2,
                        }
                    }
                }
            },
        ],
        "totalCount": 3,
    }
}

snapshots["test_order_by_question_answer_value[file-True-False] 1"] = {
    "allCases": {
        "edges": [
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [
                                {"node": {"fileValue": {"name": "f"}}},
                                {"node": {"fileValue": {"name": "a"}}},
                            ],
                            "totalCount": 2,
                        }
                    }
                }
            },
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [
                                {"node": {"fileValue": {"name": "d"}}},
                                {"node": {"fileValue": {"name": "c"}}},
                            ],
                            "totalCount": 2,
                        }
                    }
                }
            },
            {
                "node": {
                    "document": {
                        "answers": {
                            "edges": [
                                {"node": {"fileValue": {"name": "b"}}},
                                {"node": {"fileValue": {"name": "e"}}},
                            ],
                            "totalCount": 2,
                        }
                    }
                }
            },
        ],
        "totalCount": 3,
    }
}



from datetime import timedelta

import minio
from django.conf import settings


class Minio:
    def __init__(self):
        endpoint = settings.MINIO_STORAGE_ENDPOINT
        access_key = settings.MINIO_STORAGE_ACCESS_KEY
        secret_key = settings.MINIO_STORAGE_SECRET_KEY
        secure = settings.MINIO_STORAGE_USE_HTTPS
        self.client = minio.Minio(
            endpoint, access_key=access_key, secret_key=secret_key, secure=secure
        )
        self.bucket = settings.MINIO_STORAGE_MEDIA_BUCKET_NAME

    def stat_object(self, object_name):
        """
        Get stat of object in bucket.

        :param object_name: str
        :return: stat response if successful, otherwise None
        """
        try:
            return self.client.stat_object(self.bucket, object_name)
        except minio.error.NoSuchKey:  # pragma: no cover
            # object does not exist
            pass
        except minio.error.NoSuchBucket:  # pragma: no cover
            if settings.MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET:
                self.client.make_bucket(self.bucket)
                return self.stat_object(object_name)
        except minio.error.ResponseError:  # pragma: no cover
            pass

    def download_url(self, object_name):
        try:
            return self.client.presigned_get_object(
                self.bucket,
                object_name,
                timedelta(minutes=settings.MINIO_PRESIGNED_TTL_MINUTES),
            )
        except minio.error.NoSuchBucket:  # pragma: no cover
            if settings.MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET:
                self.client.make_bucket(self.bucket)
                return self.download_url(object_name)

    def upload_url(self, object_name):
        try:
            return self.client.presigned_put_object(
                self.bucket,
                object_name,
                timedelta(minutes=settings.MINIO_PRESIGNED_TTL_MINUTES),
            )
        except minio.error.NoSuchBucket:  # pragma: no cover
            if settings.MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET:
                self.client.make_bucket(self.bucket)
                return self.upload_url(object_name)

    def remove_object(self, object_name):
        self.client.remove_object(self.bucket, object_name)

    def copy_object(self, object_name, new_object_name):
        self.client.copy_object(
            self.bucket, new_object_name, f"/{self.bucket}/{object_name}"
        )

    def move_object(self, object_name, new_object_name):
        self.copy_object(object_name, new_object_name)
        self.remove_object(object_name)


if settings.MEDIA_STORAGE_SERVICE == "minio":
    client = Minio()
else:  # pragma: no cover
    client = None
    raise NotImplementedError(
        f"Storage service {settings.MEDIA_STORAGE_SERVICE} is not implemented!"
    )



from django.dispatch import Signal

created_work_item = Signal(providing_args=["work_item"])
cancelled_work_item = Signal(providing_args=["work_item"])
completed_work_item = Signal(providing_args=["work_item"])
skipped_work_item = Signal(providing_args=["work_item"])
completed_case = Signal(providing_args=["case"])
created_case = Signal(providing_args=["case"])
cancelled_case = Signal(providing_args=["case"])



import pytest


@pytest.mark.parametrize(
    "filter,expected_result",
    [
        ([{"createdBefore": "2020-04-21T00:00:00Z"}], [20]),
        ([{"createdBefore": "2020-04-22T00:00:00Z"}], [20, 21]),
        ([{"createdAfter": "2020-04-21T00:00:00Z"}], [21, 22, 23, 24]),
        ([{"createdAfter": "2020-04-22T00:00:00Z"}], [22, 23, 24]),
        # combinations
        (
            [
                {"createdBefore": "2020-04-23T00:00:00Z"},
                {"createdAfter": "2020-04-22T00:00:00Z"},
            ],
            [22],
        ),
        (
            [
                {"createdBefore": "2020-04-24T00:00:00Z"},
                {"createdAfter": "2020-04-21T00:00:00Z"},
            ],
            [21, 22, 23],
        ),
    ],
)
def test_before_after_filters(
    db, case_factory, schema_executor, filter, expected_result
):
    # thanks, auto_now_add!
    cases = case_factory.create_batch(5)
    case_dict = {}
    for day, case in zip(range(20, 25), cases):
        case_dict[day] = case.pk
        case.created_at = f"2020-04-{day} 00:00:00Z"
        case.meta = {"test_created": day}
        case.save()

    query = """
            query Q($filter: [CaseFilterSetType]) {
              allCases(filter: $filter) {
                edges {
                  node {
                    id
                    meta
                  }
                }
              }
            }
        """

    result = schema_executor(query, variable_values={"filter": filter})

    assert not result.errors
    result_info = set(
        case["node"]["meta"]["test_created"]
        for case in result.data["allCases"]["edges"]
    )

    assert result_info == set(expected_result)



import hashlib
import json

import pytest
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured
from rest_framework import status

from .. import views


@pytest.mark.parametrize("is_id_token", [True, False])
@pytest.mark.parametrize(
    "authentication_header,authenticated,error",
    [
        ("", False, False),
        ("Bearer", False, True),
        ("Bearer Too many params", False, True),
        ("Basic Auth", False, True),
        ("Bearer Token", True, False),
    ],
)
def test_authentication_view(
    rf,
    authentication_header,
    authenticated,
    error,
    is_id_token,
    requests_mock,
    settings,
):
    userinfo = {"sub": "1"}
    requests_mock.get(settings.OIDC_USERINFO_ENDPOINT, text=json.dumps(userinfo))

    if not is_id_token:
        userinfo = {"client_id": "test_client", "sub": "service-account-foo-bar"}
        requests_mock.get(settings.OIDC_USERINFO_ENDPOINT, status_code=401)
        requests_mock.post(settings.OIDC_INTROSPECT_ENDPOINT, text=json.dumps(userinfo))

    request = rf.get("/graphql", HTTP_AUTHORIZATION=authentication_header)
    response = views.AuthenticationGraphQLView.as_view()(request)
    assert bool(response.status_code == status.HTTP_401_UNAUTHORIZED) == error
    if not error:
        key = "userinfo" if is_id_token else "introspect"

        assert request.user.is_authenticated == authenticated
        if authenticated:
            assert (
                cache.get(
                    f"authentication.{key}.{hashlib.sha256(b'Token').hexdigest()}"
                )
                == userinfo
            )


@pytest.mark.parametrize("introspection", [False, True])
def test_authentication_invalid_provider(introspection, rf, requests_mock, settings):
    requests_mock.get(settings.OIDC_USERINFO_ENDPOINT, status_code=400)

    if introspection:
        requests_mock.get(settings.OIDC_USERINFO_ENDPOINT, status_code=401)
        requests_mock.post(settings.OIDC_INTROSPECT_ENDPOINT, status_code=400)
    elif not introspection:
        settings.OIDC_INTROSPECT_ENDPOINT = None

    request = rf.get("/graphql", HTTP_AUTHORIZATION="Bearer Token")
    response = views.AuthenticationGraphQLView.as_view()(request)
    assert response.status_code == status.HTTP_400_BAD_REQUEST
    result = json.loads(response.content)
    assert (
        result["errors"][0]["message"]
        == f'400 Client Error: None for url: mock://caluma.io/openid/{"introspect" if introspection else "userinfo"}'
    )


def test_authentication_view_improperly_configured(rf, settings):
    settings.OIDC_USERINFO_ENDPOINT = None
    request = rf.get("/graphql", HTTP_AUTHORIZATION="Bearer Token")
    with pytest.raises(ImproperlyConfigured):
        views.AuthenticationGraphQLView.as_view()(request)


def test_no_client_id(rf, requests_mock, settings):
    cache.clear()
    authentication_header = "Bearer Token"
    userinfo = {"sub": "1"}
    requests_mock.get(
        settings.OIDC_USERINFO_ENDPOINT, text=json.dumps(userinfo), status_code=401
    )
    requests_mock.post(settings.OIDC_INTROSPECT_ENDPOINT, text=json.dumps(userinfo))

    request = rf.get("/graphql", HTTP_AUTHORIZATION=authentication_header)
    response = views.AuthenticationGraphQLView.as_view()(request)
    assert response.status_code == status.HTTP_401_UNAUTHORIZED



import json

from graphql_relay import to_global_id

from ...caluma_form import models as form_models, schema as form_schema
from .. import permissions, types


class _TestPermission(permissions.BasePermission):
    def permission_impl(self, mutation, info):  # pragma: no cover
        raise NotImplementedError("You need to mock permission_impl")

    @permissions.permission_for(types.Node)
    def has_permission(self, mutation, info):
        # We need this indirection as this method will be registered
        # with the permissions system, but we need to mock the functionality
        # for inspection.
        return self.permission_impl(mutation, info)


def test_get_mutation_params_inline(
    db, schema_executor, permission_classes, form, mocker
):
    permission_classes([f"{__name__}._TestPermission"])

    def check_perm(self, mutation, info):
        params = mutation.get_params(info)
        if mutation == form_schema.SaveTextQuestion:
            return True
        assert len(params) == 1
        assert len(params["input"]) == 2
        assert params["input"]["form"] == form.slug
        meta = json.loads(params["input"]["meta"])
        return meta["allow"]

    mocker.patch(f"{__name__}._TestPermission.permission_impl", check_perm)

    query = r"""
        mutation foo {
            save1: saveDocument(input: {form: "%(form)s", meta: "{\"allow\": true}"})  {
              clientMutationId
            }
            save2: saveDocument(input: {form: "%(form)s", meta: "{\"allow\": false}"})  {
              clientMutationId
            }
            save3: saveTextQuestion(input: {
                slug: "some-question",
                label: "label",
                formatValidators: ["email"]
            }) {
              question {
                id
              }
            }
        }
    """ % {
        "form": form.slug
    }

    result = schema_executor(query)
    assert result.data == {
        "save1": {"clientMutationId": None},
        "save2": None,
        "save3": {
            "question": {
                "id": to_global_id(
                    "TextQuestion", form_models.Question.objects.get().slug
                )
            }
        },
    }
    assert len(result.errors) == 1


def test_get_mutation_params_with_vars(db, schema_executor, permission_classes, mocker):
    permission_classes([f"{__name__}._TestPermission"])

    def check_perm(self, mutation, info):
        params = mutation.get_params(info)
        assert len(params) == 1
        assert len(params["input"]) == 3
        assert params["input"]["slug"] == "email_addr"
        assert type(params["input"]["format_validators"]) is list
        return True

    mocker.patch(f"{__name__}._TestPermission.permission_impl", check_perm)

    query = r"""
        mutation foo ($quest: SaveTextQuestionInput!){
          saveTextQuestion(input: $quest) {
            question {
              id
            }
          }
        }
    """

    result = schema_executor(
        query,
        variable_values={
            "quest": {
                "slug": "email_addr",
                "label": "hello",
                "formatValidators": ["email"],
            }
        },
    )
    assert not result.errors

    assert result.data == {
        "saveTextQuestion": {
            "question": {
                "id": to_global_id(
                    "TextQuestion", form_models.Question.objects.get().slug
                )
            }
        }
    }



from django.core.exceptions import ValidationError
from django.db import transaction
from rest_framework import exceptions
from rest_framework.serializers import CharField, ListField
from simple_history.utils import bulk_create_with_history

from caluma.caluma_core.events import SendEventSerializerMixin

from ..caluma_core import serializers
from ..caluma_form.models import Document, Form
from . import domain_logic, events, models, utils
from .jexl import FlowJexl, GroupJexl


class FlowJexlField(serializers.JexlField):
    def __init__(self, **kwargs):
        super().__init__(FlowJexl(), **kwargs)


class GroupJexlField(serializers.JexlField):
    def __init__(self, **kwargs):
        super().__init__(GroupJexl(), **kwargs)


class SaveWorkflowSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Workflow
        fields = (
            "slug",
            "name",
            "description",
            "meta",
            "start_tasks",
            "allow_all_forms",
            "allow_forms",
            "is_archived",
            "is_published",
        )


class AddWorkflowFlowSerializer(serializers.ModelSerializer):
    workflow = serializers.GlobalIDField(source="slug")
    tasks = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=models.Task.objects, many=True
    )
    next = FlowJexlField(required=True)

    def validate_next(self, value):
        jexl = FlowJexl()
        tasks = set(jexl.extract_tasks(value))

        if not tasks:
            raise exceptions.ValidationError(
                f"jexl `{value}` does not contain any tasks as return value"
            )

        available_tasks = set(
            models.Task.objects.filter(slug__in=tasks).values_list("slug", flat=True)
        )

        not_found_tasks = tasks - available_tasks
        if not_found_tasks:
            raise exceptions.ValidationError(
                f"jexl `{value}` contains invalid tasks [{', '.join(not_found_tasks)}]"
            )

        return value

    @transaction.atomic
    def update(self, instance, validated_data):
        user = self.context["request"].user
        tasks = validated_data["tasks"]
        models.Flow.objects.filter(
            task_flows__workflow=instance, task_flows__task__in=tasks
        ).delete()
        flow = models.Flow.objects.create(
            next=validated_data["next"],
            created_by_user=user.username,
            created_by_group=user.group,
        )
        task_flows = [
            models.TaskFlow(task=task, workflow=instance, flow=flow) for task in tasks
        ]
        bulk_create_with_history(task_flows, models.TaskFlow)

        return instance

    class Meta:
        fields = ("workflow", "tasks", "next")
        model = models.Workflow


class RemoveFlowSerializer(serializers.ModelSerializer):
    flow = serializers.GlobalIDField(source="id")

    def update(self, instance, validated_data):
        models.Flow.objects.filter(pk=instance.pk).delete()
        return instance

    class Meta:
        fields = ("flow",)
        model = models.Flow


class SaveTaskSerializer(serializers.ModelSerializer):
    address_groups = GroupJexlField(
        required=False,
        allow_null=True,
        help_text=models.Task._meta.get_field("address_groups").help_text,
    )

    control_groups = GroupJexlField(
        required=False,
        allow_null=True,
        help_text=models.Task._meta.get_field("control_groups").help_text,
    )

    class Meta:
        model = models.Task
        fields = (
            "slug",
            "name",
            "description",
            "meta",
            "address_groups",
            "control_groups",
            "is_archived",
            "lead_time",
            "is_multiple_instance",
        )


class SaveSimpleTaskSerializer(SaveTaskSerializer):
    def validate(self, data):
        data["type"] = models.Task.TYPE_SIMPLE
        return super().validate(data)

    class Meta(SaveTaskSerializer.Meta):
        pass


class SaveCompleteWorkflowFormTaskSerializer(SaveTaskSerializer):
    def validate(self, data):
        data["type"] = models.Task.TYPE_COMPLETE_WORKFLOW_FORM
        return super().validate(data)

    class Meta(SaveTaskSerializer.Meta):
        pass


class SaveCompleteTaskFormTaskSerializer(SaveTaskSerializer):
    form = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=Form.objects, required=True
    )

    def validate(self, data):
        data["type"] = models.Task.TYPE_COMPLETE_TASK_FORM
        return super().validate(data)

    class Meta(SaveTaskSerializer.Meta):
        fields = SaveTaskSerializer.Meta.fields + ("form",)


class CaseSerializer(SendEventSerializerMixin, serializers.ModelSerializer):
    workflow = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=models.Workflow.objects.prefetch_related("start_tasks")
    )
    parent_work_item = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=models.WorkItem.objects, required=False
    )
    form = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=Form.objects, required=False
    )

    def validate(self, data):
        try:
            data = domain_logic.StartCaseLogic.validate_for_start(data)
        except ValidationError as e:
            raise exceptions.ValidationError(str(e))

        return super().validate(data)

    @transaction.atomic
    def create(self, validated_data):
        user = self.context["request"].user

        validated_data = domain_logic.StartCaseLogic.pre_start(validated_data, user)

        case = super().create(validated_data)

        return domain_logic.StartCaseLogic.post_start(
            case, user, validated_data.get("parent_work_item")
        )

    class Meta:
        model = models.Case
        fields = ("workflow", "meta", "parent_work_item", "form")


class SaveCaseSerializer(CaseSerializer):
    def create(self, validated_data):  # pragma: no cover
        instance = super().create(validated_data)
        self.send_event(events.created_case, case=instance)
        return instance

    def update(self, instance, validated_data):
        instance = super().update(instance, validated_data)
        self.send_event(events.created_case, case=instance)
        return instance

    class Meta(CaseSerializer.Meta):
        fields = ("id", "workflow", "meta", "parent_work_item", "form")


class CancelCaseSerializer(SendEventSerializerMixin, serializers.ModelSerializer):
    id = serializers.GlobalIDField()

    class Meta:
        model = models.Case
        fields = ("id",)

    def validate(self, data):
        try:
            domain_logic.CancelCaseLogic.validate_for_cancel(self.instance)
        except ValidationError as e:
            raise exceptions.ValidationError(str(e))

        return super().validate(data)

    @transaction.atomic
    def update(self, instance, validated_data):
        user = self.context["request"].user

        super().update(
            instance, domain_logic.CancelCaseLogic.pre_cancel(validated_data, user)
        )

        domain_logic.CancelCaseLogic.post_cancel(instance, user)

        return instance


class CompleteWorkItemSerializer(serializers.ModelSerializer):
    id = serializers.GlobalIDField()

    def validate(self, data):
        try:
            domain_logic.CompleteWorkItemLogic.validate_for_complete(
                self.instance, self.context["request"].user
            )
        except ValidationError as e:
            raise exceptions.ValidationError(str(e))

        return super().validate(data)

    @transaction.atomic
    def update(self, work_item, validated_data):
        user = self.context["request"].user

        validated_data = domain_logic.CompleteWorkItemLogic.pre_complete(
            validated_data, user
        )

        work_item = super().update(work_item, validated_data)
        work_item = domain_logic.CompleteWorkItemLogic.post_complete(work_item, user)

        return work_item

    class Meta:
        model = models.WorkItem
        fields = ("id",)


class SkipWorkItemSerializer(serializers.ModelSerializer):
    id = serializers.GlobalIDField()

    def validate(self, data):
        try:
            domain_logic.SkipWorkItemLogic.validate_for_skip(self.instance)
        except ValidationError as e:
            raise exceptions.ValidationError(str(e))

        return data

    def update(self, work_item, validated_data):
        user = self.context["request"].user

        validated_data = domain_logic.SkipWorkItemLogic.pre_skip(validated_data, user)

        work_item = super().update(work_item, validated_data)
        work_item = domain_logic.SkipWorkItemLogic.post_skip(work_item, user)

        return work_item

    class Meta:
        model = models.WorkItem
        fields = ("id",)


class SaveWorkItemSerializer(SendEventSerializerMixin, serializers.ModelSerializer):
    work_item = serializers.GlobalIDField(source="id")
    name = CharField(
        required=False,
        allow_blank=True,
        allow_null=True,
        help_text=models.WorkItem._meta.get_field("name").help_text,
    )
    description = CharField(
        required=False,
        allow_blank=True,
        allow_null=True,
        help_text=models.WorkItem._meta.get_field("description").help_text,
    )

    def update(self, instance, validated_data):
        instance = super().update(instance, validated_data)
        self.send_event(events.created_work_item, work_item=instance)
        return instance

    class Meta:
        model = models.WorkItem
        fields = (
            "work_item",
            "name",
            "description",
            "assigned_users",
            "deadline",
            "meta",
        )


class CreateWorkItemSerializer(SendEventSerializerMixin, serializers.ModelSerializer):
    case = serializers.GlobalIDPrimaryKeyRelatedField(queryset=models.Case.objects)
    multiple_instance_task = serializers.GlobalIDPrimaryKeyRelatedField(
        queryset=models.Task.objects, source="task"
    )
    controlling_groups = ListField(child=CharField(required=False), required=False)
    addressed_groups = ListField(child=CharField(required=False), required=False)
    name = CharField(
        required=False,
        allow_blank=True,
        help_text=models.WorkItem._meta.get_field("name").help_text,
    )
    description = CharField(
        required=False,
        allow_blank=True,
        help_text=models.WorkItem._meta.get_field("description").help_text,
    )

    def validate_multiple_instance_task(self, task):
        if not task.is_multiple_instance:
            raise exceptions.ValidationError(
                f"The given task type {task.type} does not allow creating multiple instances of it. Please set `isMultipleInstance` to true."
            )
        return task

    def validate(self, data):
        user = self.context["request"].user
        case = data["case"]
        task = data["task"]

        if not case.work_items.filter(
            task=task, status=models.WorkItem.STATUS_READY
        ).exists():
            raise exceptions.ValidationError(
                f"The given case {case.pk} does not have any running work items corresponding to the task {task.pk}. A new instance of a `MultipleInstanceTask` can only be created when there is at least one running sibling work item."
            )

        data["document"] = Document.objects.create_document_for_task(task, user)
        data["status"] = models.WorkItem.STATUS_READY

        if "controlling_groups" not in data:
            controlling_groups = utils.get_jexl_groups(
                task.control_groups, task, case, user
            )
            if controlling_groups is not None:
                data["controlling_groups"] = controlling_groups

        if "addressed_groups" not in data:
            addressed_groups = utils.get_jexl_groups(
                task.address_groups, task, case, user
            )
            if addressed_groups is not None:
                data["addressed_groups"] = addressed_groups

        return super().validate(data)

    def create(self, validated_data):
        instance = super().create(validated_data)
        self.send_event(events.created_work_item, work_item=instance)
        return instance

    class Meta:
        model = models.WorkItem
        fields = (
            "case",
            "multiple_instance_task",
            "name",
            "description",
            "assigned_users",
            "addressed_groups",
            "controlling_groups",
            "deadline",
            "meta",
        )
